from cilantro_ee.utils.test.utils import *
from cilantro_ee.constants.system_config import *
from cilantro_ee.core.crypto import wallet


STU = ('db929395f15937f023b4682995634b9dc19b1a2b32799f1f67d6f080b742cdb1',
       '324ee2e3544a8853a3c5a0ef0946b929aa488cbe7e7ee31a0fef9585ce398502')
DAVIS = ('21fee38471799f8c2989dd81c6d46f6c2e2db6caf63efa98a093fcba064a4b62',
         'a103715914a7aae8dd8fddba945ab63a169dfe6e37f79b4a58bcf85bfd681694')
COLIN = ('9decc7f7f0b5a4fc87ab5ce700e2d6c5d51b7565923d50ea13cbf78031bb3acf',
         '20da05fdba92449732b3871cc542a058075446fedb41430ee882e99f9091cc4d')
FALCON = ('bac886e7c6e4a9fae572e170adb333b27b590157409e62d88cc0c7bc9a7b3631',
          'ed19061921c593a9d16875ca660b57aa5e45c811c8cf7af0cfcbd23faa52cbcd')
TJ = ('cf67a180f9578afa5fd704cea39b450c1542755d73614f6a4f41b627190b83bb',
      'cb9bfd4b57b243248796e9eb90bc4f0053d78f06ce68573e0fdca422f54bb0d2')
RAGHU = ('b44a8cc3dcadbdb3352ea046ec85cd0f6e8e3f584e3d6eb3bd10e142d84a9668',
         'c1f845ad8967b93092d59e4ef56aef3eba49c33079119b9c856a5354e9ccdf84')

COOL_KIDS = [STU, DAVIS, COLIN, FALCON, TJ, RAGHU]
GENERAL_WALLETS = []

if SHOULD_MINT_WALLET:
    for i in range(NUM_WALLETS_TO_MINT):
        sk, vk = wallet.new(int_to_padded_bytes(i))
        GENERAL_WALLETS.append((sk, vk))

#ALL_WALLETS = COOL_KIDS + GENERAL_WALLETS
ALL_WALLETS = COOL_KIDS



from unittest import TestCase
from cilantro_ee.services.storage.master import MasterStorage
from cilantro_ee.core.crypto import wallet


class TestMasterDatabase(TestCase):
    def setUp(self):
        self.sk, self.vk = wallet.new()
        self.db = MasterStorage()

    def tearDown(self):
        self.db.drop_collections()

    def test_init_masterdatabase(self):
        self.assertIsNotNone(self.db.blocks)

    def test_insert_block_type_error_if_not_dict(self):
        with self.assertRaises(TypeError):
            self.db.put(123)

    def test_insert_block_returns_true_when_provided_correct_arguments(self):

        block = {
            'blockNum': 1,
            'sender': 'stu',
            'amount': 1000000
        }

        result = self.db.put(block)
        self.assertTrue(result)

    def test_get_block_number_returns_data(self):
        block = {
            'blockNum': 1,
            'sender': 'stu',
            'amount': 1000000
        }

        self.db.put(block)

        block = self.db.get_block(1)
        self.assertTrue(block)

    def test_drop_db(self):
        block = {
            'blockNum': 1,
            'sender': 'stu',
            'amount': 1000000
        }

        result = self.db.put(block)
        self.assertTrue(result)

        block = self.db.get_block(1)
        self.assertTrue(block)

        self.db.drop_collections()

        block = self.db.get_block(1)
        self.assertIsNone(block)

    def test_get_block_by_hash(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'sender': 'stu',
            'amount': 1000000
        }

        result = self.db.put(block)
        self.assertTrue(result)

        stored_block = self.db.get_block('a')

        del block['_id']

        self.assertEqual(block, stored_block)

    def test_get_block_by_number(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'sender': 'stu',
            'amount': 1000000
        }

        result = self.db.put(block)
        self.assertTrue(result)

        stored_block = self.db.get_block(1)

        del block['_id']

        self.assertEqual(block, stored_block)

    def test_fail_get_block_by_hash(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'sender': 'stu',
            'amount': 1000000
        }

        result = self.db.put(block)
        self.assertTrue(result)

        stored_block = self.db.get_block('b')

        del block['_id']

        self.assertIsNone(stored_block)

    def test_fail_get_block_by_number(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'sender': 'stu',
            'amount': 1000000
        }

        result = self.db.put(block)
        self.assertTrue(result)

        stored_block = self.db.get_block(2)

        del block['_id']

        self.assertIsNone(stored_block)

    def test_get_block_owners_block_number(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        result = self.db.indexes.collection.insert_one(block)
        self.assertTrue(result)

        owners = self.db.get_owners(1)

        self.assertEqual(block['blockOwners'], owners)

    def test_get_block_owners_by_block_hash(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        result = self.db.indexes.collection.insert_one(block)
        self.assertTrue(result)

        owners = self.db.get_owners('a')

        self.assertEqual(block['blockOwners'], owners)

    def test_get_block_owners_by_block_number_non_existent(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'sender': 'stu',
            'amount': 1000000,
            'blockOwners': ['stu', 'raghu']
        }

        result = self.db.put(block)
        self.assertTrue(result)

        owners = self.db.get_owners(2)

        self.assertIsNone(owners)

    def test_get_block_owners_by_block_hash_non_existent(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'sender': 'stu',
            'amount': 1000000,
            'blockOwners': ['stu', 'raghu']
        }

        result = self.db.put(block)
        self.assertTrue(result)

        owners = self.db.get_owners('x')

        self.assertIsNone(owners)

    def test_get_block_owners_bad_dict_returns_none(self):
        block = {
            'blockNum': 1,
            'blockHash': 'a',
            'sender': 'stu',
            'amount': 1000000,
            'blockOwners': ['stu', 'raghu']
        }

        result = self.db.put(block)
        self.assertTrue(result)

        owners = self.db.get_owners(1000)

        self.assertIsNone(owners)

    def test_create_genesis_block(self):
        block = self.db.get_block(0)

        self.assertIsNotNone(block)

    def test_query_last_n_blocks(self):
        block_1 = {
            'blockNum': 1,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        block_2 = {
            'blockNum': 2,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        block_3 = {
            'blockNum': 3,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        block_4 = {
            'blockNum': 4,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        block_5 = {
            'blockNum': 5,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        self.db.indexes.collection.insert_one(block_1)
        self.db.indexes.collection.insert_one(block_2)
        self.db.indexes.collection.insert_one(block_3)
        self.db.indexes.collection.insert_one(block_4)
        self.db.indexes.collection.insert_one(block_5)

        blocks = self.db.get_last_n(3)

        nums = [block['blockNum'] for block in blocks]

        self.assertEqual(nums, [5, 4, 3])

    def test_query_too_many_blocks_returns_total(self):
        block_1 = {
            'blockNum': 1,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        block_2 = {
            'blockNum': 2,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        block_3 = {
            'blockNum': 3,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        block_4 = {
            'blockNum': 4,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        block_5 = {
            'blockNum': 5,
            'blockHash': 'a',
            'blockOwners': ['stu', 'raghu']
        }

        self.db.indexes.collection.insert_one(block_1)
        self.db.indexes.collection.insert_one(block_2)
        self.db.indexes.collection.insert_one(block_3)
        self.db.indexes.collection.insert_one(block_4)
        self.db.indexes.collection.insert_one(block_5)

        blocks = self.db.get_last_n(300)

        nums = [block['blockNum'] for block in blocks]

        self.assertEqual(nums, [5, 4, 3, 2, 1, 0])



from unittest import TestCase
from cilantro_ee.nodes.delegate.block_manager import SubBlocks, NextBlockData, NextBlock
from cilantro_ee.constants.system_config import *
from collections import namedtuple


class TestSubBlocks(TestCase):
    def test_reset_removes_sbs_and_futures(self):
        s = SubBlocks()
        s.sbs = {1, 2, 3}
        s.futures = [1, 2, 3]

        s.reset()

        self.assertEqual(s.sbs, {})
        self.assertEqual(s.futures, [])

    def test_is_quorum_returns_true_if_sbs_equal_num_sb_per_block(self):
        s = SubBlocks()

        s.sbs = {1, 2, 3}
        self.assertTrue(s.is_quorum())

    def test_is_quorum_returns_false_if_sbs_ne_num_sb_per_block(self):
        s = SubBlocks()

        s.sbs = {1, 2}
        self.assertFalse(s.is_quorum())

    def test_add_sub_block_adds_to_proper_index(self):
        MockSB = namedtuple('MockSB', ['subBlockIdx'])

        sb = MockSB(0)

        s = SubBlocks()

        s.add_sub_block(sb, 'something')

        self.assertEqual(s.sbs[0], sb)
        self.assertListEqual(s.futures, ['something'])

    def test_get_sb_hashes_sorted(self):
        MockSB = namedtuple('MockSB', ['resultHash'])

        expected = [6, 7, 8]

        s = SubBlocks()

        s.sbs = [MockSB(6), MockSB(7), MockSB(8)]

        self.assertListEqual(s.get_sb_hashes_sorted(), expected)

    def test_add_and_get_behave_as_expected(self):
        MockSB = namedtuple('MockSB', ['subBlockIdx', 'resultHash'])

        expected = [6, 7, 8]

        s = SubBlocks()

        s.add_sub_block(MockSB(0, 6), 1)
        s.add_sub_block(MockSB(1, 7), 2)
        s.add_sub_block(MockSB(2, 8), 3)

        self.assertListEqual(s.get_sb_hashes_sorted(), expected)

    def test_get_input_hashes_sorted(self):
        MockSB = namedtuple('MockSB', ['inputHash'])

        expected = ['00', '01', '02']

        s = SubBlocks()

        s.sbs = [MockSB(b'\x00'), MockSB(b'\x01'), MockSB(b'\x02')]

        self.assertListEqual(s.get_input_hashes_sorted(), expected)

    def test_get_input_hashes_sorted_works_if_hashes_added(self):
        MockSB = namedtuple('MockSB', ['subBlockIdx', 'inputHash'])

        expected = ['00', '01', '02']

        s = SubBlocks()

        s.add_sub_block(MockSB(0, b'\x00'), 1)
        s.add_sub_block(MockSB(1, b'\x01'), 2)
        s.add_sub_block(MockSB(2, b'\x02'), 3)

        self.assertListEqual(s.get_input_hashes_sorted(), expected)


class MockBlock:
    def which(self):
        return 'SomethingElse'


class MockFailedBlock:
    def which(self):
        return 'FailedBlock'


class TestNextBlockData(TestCase):
    def test_init_non_failed_block_inits_quorum_num(self):
        n = NextBlockData(MockBlock())

        self.assertEqual(n.quorum_num, BLOCK_NOTIFICATION_QUORUM)

    def test_init_failed_block_inits_quorum_num(self):
        n = NextBlockData(MockFailedBlock())

        self.assertEqual(n.quorum_num, FAILED_BLOCK_NOTIFICATION_QUORUM)

    def test_add_sender_returns_false_if_is_quorum(self):
        n = NextBlockData(MockBlock())

        n.is_quorum = True

        self.assertFalse(n.add_sender(1))

    def test_add_sender_returns_false_if_senders_lte_quorum_num(self):
        n = NextBlockData(MockBlock())

        self.assertFalse(n.add_sender(1))

    def test_add_sender_returns_true_if_not_quorum_and_senders_gte_quorum_num(self):
        n = NextBlockData(MockBlock())

        self.assertFalse(n.add_sender(1))
        self.assertTrue(n.add_sender(2))

    def test_add_sender_returns_false_if_quorum_after_adding_senders(self):
        n = NextBlockData(MockBlock())

        self.assertFalse(n.add_sender(1))
        self.assertTrue(n.add_sender(2))
        self.assertFalse(n.add_sender(3))


class TestNextBlock(TestCase):
    def test_hard_reset_resets_state(self):
        n = NextBlock()
        n.next_block_data = {1, 2, 3}
        n.quorum_block = 1
        n.hard_reset()

        self.assertEqual(n.next_block_data, {})
        self.assertEqual(n.quorum_block, None)

    def test_reset_greater_than_existing_block_removes_block_from_set(self):
        n = NextBlock()

        MockBlock = namedtuple('MockBlock', ['blockNum'])

        n.quorum_block = MockBlock(0)
        n.next_block_data[0] = 123

        n.reset(1)

        self.assertIsNone(n.quorum_block)
        self.assertIsNone(n.next_block_data.get(0))

    def test_reset_less_than_existing_block_does_not_remove_next_block_but_removes_quorum_block(self):
        n = NextBlock()

        MockBlock = namedtuple('MockBlock', ['blockNum'])

        n.quorum_block = MockBlock(0)
        n.next_block_data[0] = 123

        n.reset(0)

        self.assertIsNone(n.quorum_block)
        self.assertEqual(n.next_block_data.get(0), 123)

    def test_is_quorum_returns_true_if_quorum_block(self):
        n = NextBlock()

        n.quorum_block = 1

        self.assertTrue(n.is_quorum())

    def test_is_quorum_returns_false_if_not_quorum_block(self):
        n = NextBlock()

        self.assertFalse(n.is_quorum())

    def test_add_notification_returns_false_if_quorum_block_and_block_num_matches_arg(self):
        n = NextBlock()

        MockBlock = namedtuple('MockBlock', ['blockNum'])

        n.quorum_block = MockBlock(999)

        self.assertFalse(n.add_notification(None, None, 999, None))

    def test_add_notification_makes_empty_hash_if_block_num_not_in_next_block_data_and_adds_data(self):
        n = NextBlock()

        self.assertIsNone(n.next_block_data.get(999))

        m = MockBlock()

        n.add_notification(m, None, 999, None)

        self.assertEqual(n.next_block_data[999][None].block_notif, m)

    def test_add_notification_returns_false_if_not_add_sender(self):
        n = NextBlock()

        m = MockBlock()

        self.assertFalse(n.add_notification(m, None, 999, None))

    def test_add_notification_returns_true_if_add_sender_returns_true(self):
        n = NextBlock()

        m = MockBlock()

        n.add_notification(m, 1, 999, None)
        self.assertTrue(n.add_notification(m, 2, 999, None))

    def test_add_notification_returns_true_if_add_sender_true_and_sets_quorum_block_to_notif(self):
        n = NextBlock()

        m = MockBlock()

        n.add_notification(m, 1, 999, None)
        n.add_notification(m, 2, 999, None)

        self.assertEqual(n.quorum_block, m)


from cilantro_ee.core.logger import get_logger
import time, os


log = get_logger("MN_URL_GETTER")


def countdown(duration: int, msg: str, log=None, status_update_freq=5):
    _l = log or get_logger("Countdown")
    if duration > status_update_freq:
        num_sleeps = duration // status_update_freq

        for _ in range(num_sleeps):
            time.sleep(status_update_freq)
            duration -= status_update_freq
            _l.important3(msg.format(duration))

    if duration > 0:
        time.sleep(duration)


def int_to_padded_bytes(i: int) -> bytes:
    SIZE = 32
    s = str(i)
    assert len(s) <= SIZE, "int {} is too long!".format(s)

    padding = SIZE - len(s)
    s = '0'*padding + s
    b = s.encode()

    assert len(b) == SIZE, "{} is not size {}".format(b, SIZE)

    return s.encode()



import zmq
import zmq.asyncio
import asyncio
import json
from cilantro_ee.utils.keys import Keys
from cilantro_ee.services.overlay.interface import OverlayInterface
from cilantro_ee.constants.overlay_network import CMD_URL
from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.services.overlay.network import Network
from cilantro_ee.constants.ports import DHT_PORT, EVENT_PORT
from cilantro_ee.constants import conf
from cilantro_ee.services.storage.vkbook import PhoneBook
from cilantro_ee.core.sockets.services import _socket, SocketStruct, SocketEncoder

def no_reply(fn):
    def _no_reply(self, *args, **kwargs):
        id_frame = args[0]
        fut = asyncio.ensure_future(fn(self, *args[1:], **kwargs))
    return _no_reply


def reply(fn):
    def _reply(self, *args, **kwargs):
        id_frame = args[0]
        res = fn(self, *args[1:], **kwargs)
        self.cmd_sock.send_multipart([
            id_frame,
            json.dumps(res, cls=SocketEncoder).encode()
        ])
    return _reply


log = get_logger('Overlay.Server')


def async_reply(fn):
    def _reply(self, *args, **kwargs):
        def _done(fut):
            self.cmd_sock.send_multipart([
                id_frame,
                json.dumps(fut.result(), cls=SocketEncoder).encode()
            ])
        id_frame = args[0]
        fut = asyncio.ensure_future(fn(self, *args[1:], **kwargs))
        fut.add_done_callback(_done)

    return _reply


class OverlayServer:
    def __init__(self, sk, ctx, quorum):
        self.log = get_logger('Overlay.Server')
        self.sk = sk
        self.wallet = Wallet(seed=sk)
        self.loop = asyncio.get_event_loop()

        Keys.setup(sk_hex=self.sk)

        self.loop = asyncio.get_event_loop()
        self.ctx = ctx
        if quorum <= 0:
            self.log.critical("quorum value should be greater than 0 for overlay server to properly synchronize!")

        self.quorum = quorum

        self.supported_methods = [func for func in dir(OverlayInterface) if callable(getattr(OverlayInterface, func)) and not func.startswith("__")]

        self.cmd_sock = self.ctx.socket(zmq.ROUTER)
        self.cmd_sock.bind(CMD_URL)

        self.network_address = 'tcp://{}:{}'.format(conf.HOST_IP, DHT_PORT)

        self.network = Network(wallet=self.wallet,
                                  ctx=self.ctx,
                                  ip=conf.HOST_IP,
                                  peer_service_port=DHT_PORT,
                                  event_publisher_port=EVENT_PORT,
                                  bootnodes=conf.BOOTNODES,
                                  initial_mn_quorum=PhoneBook.num_boot_masternodes,
                                  initial_del_quorum=PhoneBook.num_boot_delegates,
                                  mn_to_find=PhoneBook.masternodes,
                                  del_to_find=PhoneBook.delegates)

    def start(self):
        self.loop.run_until_complete(asyncio.ensure_future(
            asyncio.gather(
                self.network.start(),
                self.command_listener()
            )
        ))

    async def command_listener(self):
        self.log.info('Listening for overlay commands over {}'.format(CMD_URL))
        while True:
            msg = await self.cmd_sock.recv_multipart()
            self.log.success('GOT SOMETHING: {}'.format(msg))
            self.log.debug('[Overlay] Received cmd (Proc={}): {}'.format(msg[0], msg[1:]))
            data = [b.decode() for b in msg[2:]]

            # getattr(self, msg[1].decode())(msg[0], *data)
            func = msg[1].decode()
            if func in self.supported_methods:
                getattr(self, func)(msg[0], *data)
                # self.network.func(msg[0], *data)
            else:
                raise Exception("Unsupported API call {}".format(func))


    @no_reply
    async def ready(self, *args, **kwargs):
        self.log.debugv('Overlay Client # {} ready!'.format(self.quorum))
        self.quorum = self.quorum - 1
        if self.quorum == 0:
            await self.network.bootup()

    @reply
    def invalid_api_call(self, api_call):
        self.log.info('Overlay server got unsupported api call {}'.format(api_call))
        # raghu todo create std error enums to return
        return "Unsupported API"

    def is_valid_vk(self, vk):
        return vk in PhoneBook.all

    # seems to be a reimplementation of peer services
    @async_reply
    async def get_ip_from_vk(self, event_id, vk):
        # TODO perhaps return an event instead of throwing an error in production
        self.log.info('Time to give them a vk :'.format(vk))
        if not self.is_valid_vk(vk):
            # raghu todo - create event enum / class that does this
            return {
                'event': 'invalid_vk',
                'event_id': event_id,
                'vk': vk
            }

        response = await self.network.find_node(self.network_address, vk)  # 0.0.0.0 NO PORT
        ip = response.get(vk) if response else None
        if not ip:
            return {
                'event': 'not_found',
                'event_id': event_id,
                'vk': vk
            }

        if SocketStruct.is_valid(ip):
            ip = SocketStruct.from_string(ip).id

        return {
            'event': 'got_ip',
            'event_id': event_id,
            'ip': ip,
            'vk': vk
        }

    def teardown(self):
        try:
            self.evt_sock.close()
            try:
                self.fut.set_result('done')
            except:
                self.fut.cancel()
            self.network.teardown()
            self.cmd_sock.close()
            self.log.notice('Overlay service stopped.')
        except:
            pass




# from vmnet.testcase import BaseTestCase
# from vmnet.comm import file_listener
# import unittest, time, random, vmnet, cilantro, asyncio, ujson as json, os
# import zmq, zmq.asyncio
# from os.path import join, dirname
# from cilantro_ee.utils.test.mp_test_case import vmnet_test, wrap_func
# from cilantro_ee.core.logger.base import get_logger
#
# def run_node(node_type, idx, addr_idxs):
#     # NOTE: addr_idxs represents what each node has already discovered
#     from vmnet.comm import send_to_file
#     from cilantro_ee.services.overlay.kademlia.network import Network
#     from cilantro_ee.services.overlay.kademlia.utils import digest
#     from cilantro_ee.services.overlay.kademlia.node import Node
#     from cilantro_ee.constants.ports import DHT_PORT
#     from cilantro_ee.constants.overlay_network import MIN_DISCOVERY_NODES
#     from cilantro_ee.utils.keys import Keys
#     import asyncio, os, ujson as json
#     import zmq, zmq.asyncio
#     from os import getenv as env
#     from cilantro_ee.services.storage.vkbook import VKBook
#     VKBook.setup()
#
#     # NOTE: This logic should be improved. There are no utilities for this right now
#     all_ips = env('NODE').split(',')
#     node_objs = []
#     for nt in VKBook.constitution:
#         if nt in ('masternodes', 'delegates'):
#             for creds in VKBook.constitution[nt]:
#                 vk = creds['vk']
#                 node_id = digest(vk)
#                 ip = all_ips[len(node_objs)]
#                 node_objs.append(Node(node_id=node_id, vk=vk, ip=ip, port=DHT_PORT))
#
#     addrs = [node_objs[i] for i in addr_idxs]
#
#     async def check_nodes():
#         await asyncio.sleep(1)
#         await n.bootstrap(addrs)
#         while len(n.routing_table.get_my_neighbors()) < MIN_DISCOVERY_NODES:
#             await asyncio.sleep(2)
#         if len(n.routing_table.get_my_neighbors()) >= MIN_DISCOVERY_NODES:
#             send_to_file(env('HOST_NAME'))
#
#     from cilantro_ee.core.logger import get_logger
#     log = get_logger('{}_{}'.format(node_type, idx))
#     loop = asyncio.new_event_loop()
#     asyncio.set_event_loop(loop)
#     ctx = zmq.asyncio.Context()
#     Keys.setup(VKBook.constitution[node_type][idx]['sk'])
#     log.test('Starting {}_{}'.format(node_type, idx))
#     n = Network(Keys.vk, ctx)
#     n.tasks.clear()
#     n.tasks += [
#         n.process_requests(),
#         check_nodes()
#     ]
#     n.start()
#
# class TestBootstrap(BaseTestCase):
#     log = get_logger(__name__)
#     config_file = join(dirname(cilantro.__path__[0]), 'vmnet_configs', 'cilantro-nodes-4.json')
#     environment = {'CONSTITUTION_FILE': '2-2-2.json'}
#     enable_ui = False
#
# ################################################################################
# #   Test Setup
# ################################################################################
#
#     def setUp(self):
#         super().setUp()
#         self.all_nodes = set(self.groups['node'])
#         self.nodes_complete = set()
#
#     def tearDown(self):
#         file_listener(self, self.callback, self.timeout, 20)
#         super().tearDown()
#
#     def success(self):
#         self.log.success('=' * 128)
#         self.log.success('\t{} SUCCEEDED\t\t'.format(self.id()))
#         self.log.success('=' * 128)
#
#     def callback(self, data):
#         for node in data:
#             self.nodes_complete.add(node)
#         if self.nodes_complete == self.all_nodes:
#             self.success()
#             self.end_test()
#
#     def timeout(self):
#         self.assertEqual(self.nodes_complete, self.all_nodes)
#
# ################################################################################
# #   Tests
# ################################################################################
#
#     def test_regular_all_masters(self):
#         self.execute_python('node_1', wrap_func(run_node, 'masternodes', 0, [1]))
#         self.execute_python('node_2', wrap_func(run_node, 'masternodes', 1, [0,1]))
#         self.execute_python('node_3', wrap_func(run_node, 'delegates', 0, [0,1,2]))
#         self.execute_python('node_4', wrap_func(run_node, 'delegates', 1, [0,1,2,3]))
#         time.sleep(10)
#
#     def test_regular_one_master(self):
#         self.all_nodes.remove('node_2')
#         self.execute_python('node_1', wrap_func(run_node, 'masternodes', 0, [0]))
#         self.execute_python('node_3', wrap_func(run_node, 'delegates', 0, [0,2]))
#         self.execute_python('node_4', wrap_func(run_node, 'delegates', 1, [0,2,3]))
#
#     def test_late_delegate(self):
#         self.execute_python('node_1', wrap_func(run_node, 'masternodes', 0, [0]))
#         self.execute_python('node_2', wrap_func(run_node, 'masternodes', 1, [0,1]))
#         self.execute_python('node_3', wrap_func(run_node, 'delegates', 0, [0,1,2]))
#         self.log.test('Waiting 10 seconds before spinning up last node')
#         time.sleep(10)
#         self.execute_python('node_4', wrap_func(run_node, 'delegates', 1, [0,1,2,3]))
#
#     def test_one_master_late_delegate(self):
#         self.all_nodes.remove('node_2')
#         self.execute_python('node_1', wrap_func(run_node, 'masternodes', 0, [0]))
#         self.execute_python('node_3', wrap_func(run_node, 'delegates', 0, [0,2]))
#         self.log.test('Waiting 10 seconds before spinning up last node')
#         time.sleep(10)
#         self.execute_python('node_4', wrap_func(run_node, 'delegates', 1, [0,2,3]))
#
#     def test_one_master_down_up(self):
#         self.all_nodes.remove('node_2')
#         self.execute_python('node_1', wrap_func(run_node, 'masternodes', 0, [0]))
#         self.execute_python('node_3', wrap_func(run_node, 'delegates', 0, [0,2]))
#         self.kill_node('node_1')
#         self.execute_python('node_4', wrap_func(run_node, 'delegates', 1, [2,3]))
#         self.log.test('Waiting 10 seconds before spinning up master node again')
#         time.sleep(10)
#         self.start_node('node_1')
#         self.execute_python('node_1', wrap_func(run_node, 'masternodes', 0, [0,2,3]))
#
#     def test_one_late_masternode(self):
#         self.all_nodes.remove('node_2')
#         self.execute_python('node_3', wrap_func(run_node, 'delegates', 0, [2]))
#         self.execute_python('node_4', wrap_func(run_node, 'delegates', 1, [2,3]))
#         self.log.test('Waiting 10 seconds before spinning up master node')
#         time.sleep(10)
#         self.execute_python('node_1', wrap_func(run_node, 'masternodes', 0, [0,2,3]))
#
#     def test_race_conditions(self):
#         self.execute_python('node_1', wrap_func(run_node, 'masternodes', 0, [0]))
#         self.execute_python('node_2', wrap_func(run_node, 'masternodes', 1, [1]))
#         self.execute_python('node_3', wrap_func(run_node, 'delegates', 0, [0,2]))
#         self.execute_python('node_4', wrap_func(run_node, 'delegates', 1, [1,3]))
#
# if __name__ == '__main__':
#     unittest.main()



"""
    SBBuilderManager  (main process of delegate)

    This is the main workhorse for managing inter node communication as well as
    coordinating the sub-block production and block notifications.
    It creates sub-block builder processes to manage the parallel building of different sub-blocks.
    It publishes those sub-blocks to masters so they can assemble the new blocks
       and manages the block notifications arising from that process and the follow up actions.

"""

from cilantro_ee.core.logger.base import get_logger

from cilantro_ee.nodes.catchup import CatchupManager
from cilantro_ee.nodes.delegate.sub_block_builder import SubBlockBuilder

from cilantro_ee.services.storage.state import MetaDataStorage

from cilantro_ee.core.utils.worker import Worker

from cilantro_ee.utils.lprocess import LProcess

from cilantro_ee.constants.system_config import *
from cilantro_ee.constants.zmq_filters import DEFAULT_FILTER, NEW_BLK_NOTIF_FILTER
from cilantro_ee.constants.ports import *
from cilantro_ee.constants import conf

from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message

from cilantro_ee.core.crypto.wallet import _verify
from cilantro_ee.contracts.sync import sync_genesis_contracts
import hashlib
import asyncio, zmq, time, random
import os


# class to keep track of sub-blocks sent over from my sub-block builders
class SubBlockHandler:
    def __init__(self, num_sb_per_block=NUM_SB_PER_BLOCK):
        self.num_sb_per_block = num_sb_per_block
        self.sbs = {}

    def reset(self):
        self.sbs = {}

    def is_quorum(self):
        return len(self.sbs) == self.num_sb_per_block

    def add_sub_block(self, sub_block):
        sb_idx = sub_block.subBlockIdx % self.num_sb_per_block
        if sb_idx in self.sbs:
            # todo log it as an issue
            pass
        self.sbs[sb_idx] = sub_block

    def get_sb_hashes_sorted(self):
        sb_hashes = []
        for i in range(self.num_sb_per_block):
            sb = self.sbs[i]
            sb_hashes.append(sb.resultHash)
        return sb_hashes

    def get_input_hashes_sorted(self):
        sb_hashes = []
        num_sbs = len(self.sbs)
        for i in range(num_sbs):
            sb = self.sbs[i]
            sb_hashes.append(sb.inputHash.hex())
        return sb_hashes

    def get_new_block_hash(self, prev_block_hash):
        if not self.is_quorum():
            return 0

        h = hashlib.sha3_256()

        # first add prev block hash
        h.update(prev_block_hash)

        # add all sb result hashes in order
        for i in range(self.num_sb_per_block):
            h.update(self.sbs[i].resultHash)

        # return hash of it
        return h.digest()


class BlockNotifData:
    def __init__(self, block_notif, bn_quorum=BLOCK_NOTIFICATION_QUORUM,
                 fbn_quorum=FAILED_BLOCK_NOTIFICATION_QUORUM):
        self.block_notif = block_notif
        is_failed = block_notif.which() == "FailedBlock"
        self.quorum_num = fbn_quorum if is_failed else bn_quorum
        self.is_quorum = False
        self.senders = set()

    def is_quorum(self):
        return self.is_quorum

    def add_sender(self, sender):
        self.senders.add(sender)
        if not self.is_quorum and (len(self.senders) >= self.quorum_num):
            self.is_quorum = True
            return True
        return False

# Keeps track of block notifications from master
class BlockNotifHandler:
    def __init__(self):
        self.hard_reset()

    # use this when it has to go to catchup
    def hard_reset(self):
        self.block_notif_data = {}    # hash of block num -> block hash -> data
        self.quorum_block = None

    def reset(self, block_num):
        if self.quorum_block:
            bn = self.quorum_block.blockNum
            if bn < block_num and bn in self.block_notif_data:
                try:
                    del self.block_notif_data[bn]
                except KeyError:
                    pass
                    # todo add a debug message - not supposed to happen
        self.quorum_block = None

    def is_quorum(self):
        return self.quorum_block != None

    def get_quorum_block(self):
        return self.quorum_block

    def add_notification(self, block_notif, sender):
        block_num = block_notif.blockNum
        block_hash = block_notif.blockHash
        if self.quorum_block and (self.quorum_block.blockNum == block_num):
            # todo - if it is not matching blockhash, may need to audit it
            return False

        if block_num not in self.block_notif_data:
            self.block_notif_data[block_num] = {}
            # todo add time info to implement timeout

        if block_hash not in self.block_notif_data[block_num]:
            self.block_notif_data[block_num][block_hash] = BlockNotifData(block_notif)

        if self.block_notif_data[block_num][block_hash].add_sender(sender):
            self.quorum_block = block_notif
            return True

        return False


class DBHandler:
    def __init__(self):
        self.is_db_updated = False
        self.catchup_mgr = None

    def is_ready_for_next_sb(self):
        return self.is_db_updated

    def set_catchup_needed(self):
        self.is_db_updated = False

    def set_catchup_done(self):
        ret_code = not self.is_db_updated
        self.is_db_updated = True
        return ret_code

    def setup_catchup_mgr(self, vk, sk, pub_sock, router_sock):
        self.catchup_mgr = CatchupManager(
                                          verifying_key=vk,
                                          signing_key=sk,
                                          pub_socket=pub_sock,
                                          router_socket=router_sock,
                                          store_full_blocks=False)

    def start_catchup_process(self):
        assert self.catchup_mgr, "Expected catchup_mgr initialized at this point"

        self.catchup_mgr.run_catchup()

    def recv_block_data_reply(self, reply):
        if self.catchup_mgr.recv_block_data_reply(reply) and \
           self.set_catchup_done():
            return True
        return False

    def recv_block_idx_reply(self, sender, reply):
        if self.catchup_mgr.recv_block_idx_reply(sender, reply) and \
           self.set_catchup_done():
            return True
        return False

    def recv_block_notif(self, block):
        self.set_catchup_needed()
        self.catchup_mgr.recv_new_blk_notif(block) 


class SubBlockManager:
    def __init__(self, sk, vk, sb_builder_requests,
                 min_mn_quorum, num_sb_builders=NUM_SB_BUILDERS):
        self.signing_key = sk
        self.verifying_key = vk
        self.sbb_requests = sb_builder_requests
        self.num_sb_builders = num_sb_builders
        self.sb_handler = SubBlockHandler()
        self.bn_handler = BlockNotifHandler()
        self.db_handler = DBHandler()
        self.driver = MetaDataStorage()


    def reset(self):
        # reset all the state info
        self.bn_handler.reset(self.driver.latest_block_num)
        self.sb_handler.reset()

    def setup_catchup_mgr(self, pub_sock, router_sock):
        self.db_handler.setup_catchup_mgr(self.verifying_key,
                                          self.signing_key,
                                          pub_sock,
                                          router_sock)

    def start_catchup_process(self):
        self.log.info("Catching up...")

        # Add genesis contracts to state db if needed
        sync_genesis_contracts()
        self.db_handler.start_catchup_process()

    # is passing callbacks better way
    def recv_block_data_reply(self, reply):
        if self.db_handler.recv_block_data_reply(reply):
            self.make_next_sb()

    # is passing callbacks better way
    def recv_block_idx_reply(self, sender, reply):
        if self.db_handler.recv_block_idx_reply(sender, reply):
            self.make_next_sb()

    # todo - make sure it is called when only current state is behind this block
    def recv_block_notif(self, block):
        self.db_handler.recv_block_notif(block)



    # rpc todo
    # seems like driver goes with bn-handler, perhaps at manager level which is passing
    # need to move this to blk notif handler
    # make sure block aggregator adds block_num for all notifications?
    def handle_block_notification(self, frames, block, sender: bytes):

        # todo - convert this to string hex for audit purposes ?
        self.log.notice('BM with sender {} being handled'.format(sender))
        self.log.info("Got block notification for block num {} with hash {}"
                                   .format(block.blockNum, block.blockHash))

        cur_block_num = self.driver.latest_block_num 
        next_block_num = cur_block_num + 1

        if block.blockNum < next_block_num:
            self.log.info("New block notification with block num {} that is "
                          "not more than our curr block num {}. Ignoring .."
                          .format(block.blockNum, cur_block_num))
            return

        is_quorum_met = self.bn_handler.add_notification(block, sender)

        if is_quorum_met:
            self.log.info("New block quorum met!")
            if block.blockNum > next_block_num:
                self.log.warning("Current block (# {}) is behind by more than "
                                 "one block from the received block (# {}). "
                                 "Need to run catchup!"
                                 .format(cur_block_num, block.blockNum))

            my_new_block_hash = self.sb_handler.get_new_block_hash()
            self.log.info('New hash {}, recieved hash {}'.format(my_new_block_hash, block.blockHash.hex()))

            if my_new_block_hash == block.blockHash:
                if block.which() == "newBlock":
                    self.driver.latest_block_num = block.blockNum
                    self.driver.latest_block_hash = my_new_block_hash

                    # Set the epoch hash if a new epoch has begun
                    if block.blockNum % conf.EPOCH_INTERVAL == 0:
                        self.driver.latest_epoch_hash = my_new_block_hash

                self.commit_cur_db()

            else:
                self.log.critical(
                    'BlockNotification hash received is not the same as the one we have!!!\n{}\n{}'.format(
                        my_new_block_hash, block.blockHash))

                # simply forward the block notification. it is input align on sbb
                # self.discord_cur_db(block.inputHashes)
                self.discord_cur_db(frames)
                if block.which() == "newBlock":
                    self.sb_mgr.reset()
                    self.recv_block_notif(block)
                    return
            self.make_next_sb()

    def commit_cur_db(self):
        self.sbb_requests['commit_cur_sb']()


    # def discord_cur_db(self, input_hashes):
    def discord_cur_db(self, frames):
        self.sbb_requests['discord_cur_sb_and_align'](frames)

    def make_next_sb(self):
        self.sbb_requests['make_next_sb']()


class SBBState:
    def __init__(self, num_sb_builders=NUM_SB_BUILDERS):
        self.num_sb_builders = num_sb_builders
        self.sbb_ready_count = 0

        self.pending_work_at_sbb = 0         # bit map
        
    def set_sbb_ready(self):
        self.sbb_ready_count += 1

    def are_sbbs_ready(self):
        return self.sbb_ready_count == self.num_sb_builders

    def set_pending_work(self, sbb_index):
        self._pending_work_at_sbb |= (1 << sbb_index)

    def set_no_work(self, sbb_index):
        self._pending_work_at_sbb &= ~(1 << sbb_index)

    def is_ready_for_next_sb(self):
        return (self.pending_work_at_sbb > 0)

# start catch up only when these two are true
#               (self.not_met_mn_quorum <= 0) and (self.not_ready_sbb_count <= 0)
class SubBlockBuilderManager(Worker):
    def __init__(self, ip, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = get_logger("SBBuilderManager[{}]".format(self.verifying_key[:8]))

        self.ip = ip
        self.ipc_port = 6967   # hard coded port number right now
        self.ipc_ip =  'sbb-mgr-' + str(os.getpid()) + '-' + str(random.randint(0, 2**32))

        self.mn_quorum_min = PhoneBook.masternode_quorum_min
        self.mn_ready_count = 0
        self.masternodes_ready = set()

        # only initial start up really depends on these two
        self.sb_builders = {}  # index -> process
        self.num_sb_builders = NUM_SB_BUILDERS
        self.sbb_state = SBBState(self.num_sb_builders)

        # crp todo - pass in self.log or eliminate logging at this level??
        self.sbb_requests = {
                'make_next_sb': self.send_make_next_sb,
                'commit_cur_sb': self.send_commit_cur_sb,
                'discord_cur_sb_and_align': self.send_discord_cur_sb_and_align}
        self.sb_mgr = SubBlockManager(self.signing_key, self.verifying_key,
                                      self.mn_quorum_min, self.sbb_requests)

        self._thicc_log()

        # Define Sockets (these get set in build_task_list)
        self.router, self.ipc_router, self.pub, self.sub = None, None, None, None


        self.run()

    def _thicc_log(self):
        self.log.notice("\nSBBuilderManager initializing with\nvk={vk}\n"
                        "num_sub_blocks={num_sb}\nnum_blocks={num_blocks}\nsub_blocks_per_block={sb_per_block}\n"
                        "num_sb_builders={num_sb_builders}\nsub_blocks_per_builder={sb_per_builder}\n"
                        "sub_blocks_per_block_per_builder={sb_per_block_per_builder}\n"
                        .format(vk=self.verifying_key, num_sb=NUM_SUB_BLOCKS,
                                num_blocks=NUM_BLOCKS, sb_per_block=NUM_SB_PER_BLOCK,
                                num_sb_builders=NUM_SB_BUILDERS, sb_per_builder=NUM_SB_PER_BUILDER,
                                sb_per_block_per_builder=NUM_SB_PER_BLOCK_PER_BUILDER))

    def run(self):
        # sync set up
        self.create_sockets()
        self.build_task_list()

        self.sb_mgr.setup_catchup_mgr(self.pub, self.router)

        self.log.info("Block Sub-block coordinator starting...")

        self.loop.run_until_complete(asyncio.gather(*self.tasks))


    def create_sockets(self):
        # Create a TCP Router socket for comm with other nodes
        self.router = self.manager.create_socket(
            socket_type=zmq.ROUTER,
            name="SBBM-Router-{}".format(self.verifying_key[-4:]),
            secure=True,
        )
        # self.router.setsockopt(zmq.ROUTER_MANDATORY, 1)  # FOR DEBUG ONLY
        self.router.setsockopt(zmq.IDENTITY, self.verifying_key.encode())
        self.router.bind(port=DELEGATE_ROUTER_PORT, protocol='tcp', ip=self.ip)

        # Create ROUTER socket for bidirectional communication with SBBs over IPC
        self.ipc_router = self.manager.create_socket(socket_type=zmq.ROUTER, name="SBBM-IPC-Router")
        self.ipc_router.setsockopt(zmq.ROUTER_MANDATORY, 1)  # FOR DEBUG ONLY
        self.ipc_router.bind(port=self.ipc_port, protocol='ipc', ip=self.ipc_ip)

        # Create PUB socket to publish new sub_block_contenders to all masters
        self.pub = self.manager.create_socket(
            socket_type=zmq.PUB,
            name="SBBM-Pub-{}".format(self.verifying_key[-4:]),
            secure=True,
        )
        self.pub.bind(port=DELEGATE_PUB_PORT, protocol='tcp', ip=self.ip)


        # Create SUB socket to
        # 1) listen for subblock contenders from other delegates
        # 2) listen for NewBlockNotifications from masternodes
        self.sub = self.manager.create_socket(
            socket_type=zmq.SUB,
            name="SBBM-Sub-{}".format(self.verifying_key[-4:]),
            secure=True,
        )
        self.sub.setsockopt(zmq.SUBSCRIBE, DEFAULT_FILTER.encode())
        self.sub.setsockopt(zmq.SUBSCRIBE, NEW_BLK_NOTIF_FILTER.encode())

    
    def build_task_list(self):
        self.tasks.append(self.router.add_handler(self.handle_router_msg))
        self.tasks.append(self.ipc_router.add_handler(self.handle_ipc_msg, None, True))
        self.tasks.append(self.sub.add_handler(self.handle_sub_msg))
        self.tasks.append(self.async_setup1())
        self.tasks.append(self.async_setup2())

    def set_masternode_ready(self, sender):
        if sender not in self.masternodes_ready:
            self.masternodes_ready.add(sender)
            self.mn_ready_count += 1

    def start_sbb_procs(self, ipc_ip, ipc_port, ip):
        kwargs={'ipc_ip': ipc_ip, 'ipc_port': ipc_port,
                'signing_key': self.signing_key, 'ip': self.ip}
        for i in range(self.num_sb_builders):
            kwargs['sbb_index'] = i
            self.sb_builders[i] = LProcess(target=SubBlockBuilder,
                                           name="SBB_Proc-{}".format(i),
                                           kwargs=kwargs)

            self.log.info("Starting SBB #{}".format(i))
            self.sb_builders[i].start()

    async def async_setup1(self):
        self.start_sbb_procs(self.ipc_ip, self.ipc_port, self.ip)

    async def async_setup2(self):
        # first make sure, we have overlay server ready
        await self._wait_until_ready()

        # Listen to Masternodes over sub and connect router for catchup communication
        for vk in PhoneBook.masternodes:
            self.sub.connect(vk=vk, port=MN_PUB_PORT)
            self.router.connect(vk=vk, port=MN_ROUTER_PORT)
        # let's wait a bit so connections are established properly

        # need to wait for 6 secs to let connections form as well as
        # sub-block builders and min number of master nodes ready
        wait_time = 0
        while (wait_time < 6) or not self.sbb_state.are_sbbs_ready() or \
              (self.mn_ready_count < self.mn_quorum_min): 
            wait_time += 1
            asyncio.sleep(1)

        # now start the catchup
        self.sb_mgr.start_catchup_process()


    async def handle_ipc_msg(self, frames):
        self.log.spam("Got msg over ROUTER IPC from a SBB with frames: {}".format(frames))  # TODO delete this
        assert len(frames) == 3, "Expected 3 frames: (id, msg_type, msg_blob). Got {} instead.".format(frames)

        sbb_index = int(frames[0].decode())
        self.log.info('SBBINDEX {}'.format(sbb_index))
        assert sbb_index in self.sb_builders, "Got IPC message with ID {} that is not in sb_builders {}" \
            .format(sbb_index, self.sb_builders)

        mtype_enc = frames[1]
        msg_blob = frames[2]

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message(mtype_enc, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        if msg_type == MessageType.READY:
            self.log.info('Ready signal received from a SB builder')
            self.sbb_state.set_sbb_ready()

        elif msg_type == MessageType.PENDING_TRANSACTIONS:
            self.sbb_state.set_pending_work(sbb_index)

        elif msg_type == MessageType.NO_TRANSACTIONS:
            self.sbb_state.set_no_work(sbb_index)

        elif msg_type == MessageType.SUBBLOCK_CONTENDER:
            await self._handle_sbc(sbb_index, msg, mtype_enc, msg_blob)

    def handle_sub_msg(self, frames):
        self.log.success('GOT A SUB MESSAGE ON BLOCK MGR')
        self.log.success(frames)

        # Unpack the frames
        msg_filter, msg_type, msg_blob = frames

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        # Process external ready signals
        if msg_type == MessageType.READY:
            self.set_masternode_ready(sender)

        # Process block notification messages
        elif msg_type == MessageType.BLOCK_NOTIFICATION:
            self.log.important3("Got BlockNotification from {} with hash {}"
                                .format(sender, msg.blockHash.hex()))

            # Process accordingly
            self.sb_mgr.handle_block_notification(msg, sender)


    def handle_router_msg(self, frames):
        sender, msg_type, msg_blob = frames

        msg_type, msg, signer, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, signer, timestamp))
            return

        if msg_type == MessageType.BLOCK_INDEX_REPLY:
            self.sb_mgr.recv_block_idx_reply(sender, msg)

        elif msg_type == MessageType.BLOCK_DATA:
            self.sb_mgr.recv_block_data_reply(msg)


    # todo
    # raw message -> action map and decoded message -> action maps?
    # actually add an internal function(s) to Message that can repack with or without signature ?
    async def _handle_sbc(self, sbb_index: int, sbc, mtype_enc: bytes, msg_blob: bytes):
        self.log.important("Got SBC with sb-index {} input-hash {}"
                           .format(sbc.subBlockIdx, sbc.inputHash.hex()))

        self.sb_mgr.sb_handler.add_sub_block(sbc)

        # first sign the message 
        mtype_sgn, msg_sgn = Message.get_message_signed_internal(
                                  signee=self.wallet.verifying_key(),
                                  sign=self.wallet.sign,
                                  msg_type=mtype_enc, msg=msg_blob)
        self.pub.send_msg(filter=DEFAULT_FILTER.encode(),
                          msg_type=mtype_sgn, msg=msg_sgn)

    # TODO make this DRY
    def _send_msg_over_ipc(self, sb_index: int, msg_type, message):
        """
        Convenience method to send a message over IPC router socket to a particular SBB process. Includes a
        frame to identify the type of message
        """
        self.log.spam("Sending msg to sb_index {} with payload {}".format(sb_index, message))

        id_frame = str(sb_index).encode()
        self.ipc_router.send_multipart([id_frame, msg_type, message])


    def send_ipc_message(self, msg_type: bytes, msg: bytes):
        for idx in range(self.num_sb_builders):
            self.ipc_router.send_multipart([str(idx).encode(), msg_type, msg])

    async def wait_for_work(self):
        # return immediately if at least one sb has some work, otherwise wait
        wait_time = 0
        while not self.sbb_state.is_ready_for_next_sb() and \
              wait_time < BLOCK_HEART_BEAT_INTERVAL:
            await asyncio.sleep(1)
            wait_time += 1
        if wait_time > 0:
            blk_str = "block" if wait_time < BLOCK_HEART_BEAT_INTERVAL \
                      else "empty block"
            self.log.info("Waited for {} secs to make next {}"
                          .format(wait_time, blk_str))


    def send_make_next_sb(self):
        # first wait until there is some work or timeout
        asyncio.get_event_loop().run_until_complete(self.wait_for_work())

        msg_type, msg = Message.get_message_packed(MessageType.MAKE_NEXT_SB)
        self.send_ipc_message(msg_type, msg)

    def send_commit_cur_sb(self):
        msg_type, msg = Message.get_message_packed(MessageType.COMMIT_CUR_SB)
        self.send_ipc_message(msg_type, msg)

    def xx_send_discord_cur_sb_and_align(self, input_hashes):
        for i in range(len(input_hashes)):
            idx = i % self.num_sb_builders
            mtype, msg = Message.get_message_packed(MessageType.DISCORD_AND_ALIGN,
                                                    inputHashes=input_hashes[i])
            self.ipc_router.send_multipart([str(idx).encode(), mtype, msg])

    def send_discord_cur_sb_and_align(self, frames):
        self.send_ipc_message(frames[1], frames[2])
     



from unittest import TestCase
from cilantro_ee.nodes.masternode.webserver import app, client
import json

class TestWS(TestCase):
    def test_app(self):
        _, response = app.test_client.get('/')
        self.assertEqual(response.status, 200)

    def test_ping_api(self):
        _, response = app.test_client.get('/ping')
        self.assertEqual(response.status,200)
        self.assertEqual(response.json, [True, 'Hello'])

    def test_id_api(self):
        _, response = app.test_client.get('/id')
        self.assertEqual(response.json, {'verifying_key': None})

    def test_epoch_api(self):
        _, response = app.test_client.get('/epoch')
        print(response.json)
        self.assertEqual(response.json,
                         {'epoch_hash': '0000000000000000000000000000000000000000000000000000000000000000',
                          'blocks_until_next_epoch': 1})

    def test_get_submission_methods(self):
        _, response = app.test_client.get('/contracts/submission/methods')

        method = 'submit_contract'
        kwargs = ['name', 'code', 'owner', 'constructor_args']

        methods = response.json.get('methods')

        self.assertEqual(len(methods), 1)

        test_method = methods[0]

        self.assertEqual(method, test_method.get('name'))
        self.assertListEqual(kwargs, test_method.get('arguments'))
        self.assertEqual(response.status, 200)

    def test_get_non_existent_contract_methods(self):
        _, response = app.test_client.get('/contracts/huuuuuuuuupluh/methods')

        self.assertEqual(response.status, 404)

        error_message = response.json.get('error')

        self.assertEqual(error_message, 'huuuuuuuuupluh does not exist')

    def test_contracts_api(self):
        _, response = app.test_client.get('/contracts/vkbook')
        pass

    def test_blocks_api(self):
        _, response = app.test_client.get('/latest_block')
        print(response.json)
        pass



from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('2-2-4.json')


from unittest import TestCase, mock
from unittest.mock import MagicMock
from cilantro_ee.core.utils.worker import Worker
from cilantro_ee.messages.base.base_signal import SignalBase


from cilantro_ee.constants.testnet import *
from cilantro_ee.nodes.base import NodeTypes


# A mock message type used just for these tests
class TestMessage(SignalBase):
    pass


class WorkerTester:

    @staticmethod
    def test(func):
        @mock.patch("cilantro_ee.core.utils.worker.asyncio", autospec=True)
        # @mock.patch("cilantro_ee.core.utils.worker.zmq", autospec=True)
        # @mock.patch("cilantro_ee.core.utils.worker.zmq.asyncio", autospec=True)
        @mock.patch("cilantro_ee.core.utils.worker.SocketManager")
        def _func(*args, **kwargs):
            return func(*args, **kwargs)
        return _func


class TestWorker(TestCase):

    @WorkerTester.test
    def setUp(self, *args):
        self.sk = 'A' * 64
        self.worker = Worker(signing_key=self.sk, name='TestWorker')




from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('1-0-0.json')

from vmnet.testcase import BaseNetworkTestCase
import unittest, cilantro_ee
from os.path import join, dirname
from cilantro_ee.utils.test.mp_test_case import vmnet_test
from cilantro_ee.core.logger import get_logger

LOG_LEVEL = 0


def wrap_func(fn, *args, **kwargs):
    def wrapper():
        return fn(*args, **kwargs)
    return wrapper


def run_mn():
    from cilantro_ee.utils.factory import NodeFactory
    from cilantro_ee.constants.testnet import TESTNET_MASTERNODES
    import os
    import time

    # overwrite_logger_level(logging.WARNING)
    # overwrite_logger_level(21)
    # overwrite_logger_level(11)
    print("sleep sleep before ")
    time.sleep(10)

    ip = os.getenv('HOST_IP')
    print("\n\n\ MN HAS IP {} \n\n".format(ip))
    sk = TESTNET_MASTERNODES[0]['sk']
    NodeFactory.run_masternode(ip=ip, signing_key=sk, reset_db=True)


def run_user():
    from cilantro_ee.utils.test.god import God
    from cilantro_ee.core.logger import get_logger
    from cilantro_ee.core.crypto import wallet

    # Fix masternode URL on God
    God.multi_master = False
    God.mn_urls = ['http://172.29.0.1:8080']

    log = get_logger("TestUser")
    # overwrite_logger_level(logging.WARNING)

    sk, vk = wallet.new()
    sk2, vk2 = wallet.new()

    # Request the nonce
    log.important("requesting nonce for vk: {}".format(vk))
    nonce = God.request_nonce(vk)
    log.important("got nonce return {}".format(nonce))
    assert 'success' in nonce, "'success' key not in nonce payload {}".format(nonce)
    assert 'nonce' in nonce, "'nonce' key not in nonce payload {}".format(nonce)

    log.notice("creating tx with nonce {}".format(nonce['nonce']))
    tx = God.create_currency_tx(sender=(sk, vk), receiver=(sk2, vk2), amount=10, nonce=nonce['nonce'])
    log.important("sending tx: {}".format(tx))
    r = God.send_tx(tx)

    log.important2("Got reply from sending tx: {} ... with json {}".format(r, r.json()))


class TestNonce(BaseNetworkTestCase):

    config_file = join(dirname(cilantro_ee.__path__[0]), 'vmnet_configs', 'cilantro_ee-mn.json')
    PROFILE_TYPE = None

    @vmnet_test(run_webui=True)
    def test_dump(self):
        log = get_logger("test_dump")

        log.important3("Starting masternode")
        self.execute_python('mn', wrap_func(run_mn), async=True, profiling=self.PROFILE_TYPE)

        input("Press key to start mgmt")

        log.important3("Starting mgmt")
        self.execute_python('mgmt', wrap_func(run_user), async=True, profiling=self.PROFILE_TYPE)

        input("Press any key to exit...")


if __name__ == '__main__':
    unittest.main()



# Internode filters
DEFAULT_FILTER = '0'
WITNESS_MASTERNODE_FILTER = 'wit_mn'
DELEGATE_DELEGATE_FILTER = 'del_del'
MASTERNODE_DELEGATE_FILTER = 'mn_del'
WITNESS_DELEGATE_FILTER = 'wit_del'
TRANSACTION_FILTER = 'tx'

# Catchup Filters
BLOCK_IDX_REQ_FILTER = 'blk_idx_req'

# New Block Notif Filters
NEW_BLK_NOTIF_FILTER = 'new_blk_notif'

# IPC Filters
STATESYNC_FILTER = 'state_sync'



import abc

class OverlayInterface(abc.ABC):
    def __init__(self):
        pass
        # interface dictionary

    # ready signal from client to start overlay server
    @abc.abstractmethod
    def ready(self):
        pass

    # returns ip address if found else None
    @abc.abstractmethod
    def get_ip_from_vk(self, vk):
        pass

    # returns ip address if found else None
    @abc.abstractmethod
    def get_ip_and_handshake(self, vk, is_first_time):
        pass

    @abc.abstractmethod
    def handshake_with_ip(self, vk, ip, is_first_time):
        pass

    @abc.abstractmethod
    def ping_ip(self, vk, ip, is_first_time):
        pass




import asyncio, time, cilantro_ee, os, shutil, signal, sys
import zmq.asyncio
from vmnet.launch import launch
from vmnet.webserver import start_ui
from vmnet.testcase import BaseNetworkTestCase
from cilantro_ee.core.logger.base import get_logger
from .mp_test import SIG_ABORT, SIG_FAIL, SIG_RDY, SIG_SUCC, SIG_START
from os.path import dirname, join
from functools import wraps

# URL of orchestration node. TODO -- set this to env vars
URL = "tcp://127.0.0.1:5020"

TEST_TIMEOUT = 5
TESTER_POLL_FREQ = 0.1

CILANTRO_PATH = dirname(dirname(cilantro_ee.__path__[0]))

log = get_logger("MPTestCase")  # TODO delete dat

def wrap_func(fn, *args, **kwargs):
    def wrapper():
        return fn(*args, **kwargs)
    return wrapper

def signal_handler(sig, frame):
    print("Killing docker containers...")
    os.system("docker kill $(docker ps -q)")
    print("Docker containers be ded")
    sys.exit(0)

def vmnet_test(*args, **kwargs):
    def _vmnet_test(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self = args[0]
            assert isinstance(self, BaseNetworkTestCase), \
                "@vmnet_test can only be used to decorate BaseNetworkTestCase subclass methods (got self={}, but expected " \
                "a BaseNetworkTestCase subclass instance)".format(self)

            klass = self.__class__
            config_file = klass.config_file
            parent_klass = self.__class__.__bases__[0]  # In Cilantro, this should be MPTestCase

            # Horrible hack to get MPTestCase to work
            if parent_klass is not BaseNetworkTestCase:
                klass = parent_klass

            klass.test_name = klass.__name__
            config_dict = launch(config_file, klass.test_name)
            klass._set_configs(klass, config_dict)

            # Create log directory for test name
            log_dir = join(klass.project_path, 'logs', klass.test_name)
            try: shutil.rmtree(log_dir)
            except: pass
            os.makedirs(log_dir, exist_ok=True)

            if run_webui:
                klass.webserver_proc, klass.websocket_proc = start_ui(
                    klass.test_name, klass.project_path)

            klass.vmnet_test_active = True
            res = func(*args, **kwargs)
            klass.vmnet_test_active = False

            return res

        return wrapper

    if len(args) == 1 and callable(args[0]):
        run_webui = False
        return _vmnet_test(args[0])
    else:
        run_webui = kwargs.get('run_webui', False)
        return _vmnet_test


class MPTestCase(BaseNetworkTestCase):
    config_file = '{}/cilantro_ee/vmnet_configs/cilantro_ee-nodes-4.json'.format(CILANTRO_PATH)
    testers = []
    _curr_tester_index = 1
    vmnet_test_active = False
    log_lvl = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = get_logger("MPTestOrchestrater")

    @classmethod
    def setUpClass(cls):
        MPTestCase.log_lvl = cls.log_lvl

    @classmethod
    def next_container(cls) -> tuple:
        """
        Retreives the next available docker image.
        :return: A 2 tuple containing the ip and name of container in the form: (name: str, ip: str)
        """
        num = MPTestCase._curr_tester_index
        name = "node_{}".format(num)

        assert num <= len(cls.nodemap), "Tester object number {} exceeds tester capacity of {}".format(num, len(cls.nodemap))
        assert name in cls.nodemap, "Node named {} not found in node map {}".format(name, cls.nodemap)

        MPTestCase._curr_tester_index += 1

        return name, cls.nodemap[name]

    def setUp(self):
        super().setUp()
        assert len(MPTestCase.testers) == 0, "setUp called but God._testers is not empty ({})" \
            .format(MPTestCase.testers)

        start_msg = '\n' + '#' * 80 + '\n'
        start_msg += '\n{} STARTING\n'.format(self.id()) + '\n' + '#' * 80
        self.log.test(start_msg)

    def tearDown(self):
        super().tearDown()

        MPTestCase.testers.clear()
        MPTestCase._curr_tester_index = 1

    def start(self, timeout=TEST_TIMEOUT):
        """
        Start the test orchestrator, which polls tester objects living on seperate process/machines for assertions.

        We organize all registered testers into 'active' testers, which are passed in an assertions function,
        and 'passive' testers which do not make assertions but still interact with other testers.
        Testers' output queues are polled  every TEST_POLL_FREQ seconds for maximum of TEST_TIMEOUT seconds,
        waiting for them to finish (if ever). When an 'active' testers passes its assertions, we move it to
        'passives'. When all active testers are finished, we send SIG_ABORTs to all testers to clean them up
        """
        # self.log.critical("\nSTARTING TEST WITH TESTERS {}\n".format(God.testers))
        assert len(MPTestCase.testers) > 0, "start() called, but list of testers empty (MPTesterBase._testers={})"\
                                            .format(MPTestCase.testers)

        actives, passives, fails, timeout = self._poll_testers(timeout)

        self.log.debug("Cleaning up tester processes")
        for t in actives + passives + fails:
            t.socket.send_pyobj(SIG_ABORT)
            t.teardown()

        # If there are no active testers left and none of them failed, we win
        if len(actives) + len(fails) == 0:
            self.log.test("\n{0}\n\n{2} SUCCEEDED WITH {1} SECONDS LEFT\n\n{0}"
                          .format('$' * 120, round(timeout, 2), self.id()))
        else:
            fail_msg = "\n\nfail_msg:\n{0}\nASSERTIONS TIMED OUT FOR TESTERS: \n\n".format('-' * 120)
            for t in fails + actives:
                fail_msg += "{}\n".format(t)
            fail_msg += "{0}\n".format('-' * 120)
            self.log.fatal(fail_msg)
            time.sleep(0.2)  # block while this message has time to log correctly
            raise Exception("Test(s) did not pass. See log.")

    def _poll_testers(self, timeout) -> tuple:
        start_msg = '\n' + '~' * 80
        start_msg += '\nPolling testers procs every {} seconds, with test timeout of {} seconds\n'\
            .format(TESTER_POLL_FREQ, timeout)
        start_msg += '~' * 80
        self.log.test(start_msg)

        actives = [t for t in MPTestCase.testers if t.assert_fn]
        passives = [t for t in MPTestCase.testers if not t.assert_fn]
        fails = []

        # Start the assertion on the active tester procs
        for t in actives:
            t.socket.send_pyobj(SIG_START)

        # Poll testers for a max of 'timeout' seconds
        while timeout > 0:
            for t in actives:
                try:
                    msg = t.socket.recv(flags=zmq.NOBLOCK)
                    self.log.debug("GOT MSG {} FROM TESTER <{}>".format(msg, t))

                    # 'ignore' SIG_RDY
                    if msg == SIG_RDY:
                        continue

                    actives.remove(t)

                    if msg == SIG_SUCC:
                        passives.append(t)
                    else:
                        self.log.fatal("Got non success signal {} from Tester named {}. Marking him as a failure."
                                       .format(msg, t.name))
                        fails.append(t)
                except Exception as e:
                    pass

            # Break if all active testers have finished
            if len(actives) == 0:
                self.log.debug("Breaking from loop, no active tester left")
                break

            # Sleep for TEST_POLL_FREQ seconds
            timeout -= TESTER_POLL_FREQ
            time.sleep(TESTER_POLL_FREQ)

        return actives, passives, fails, timeout



from cilantro_ee.core.utils.worker import Worker
from cilantro_ee.messages.base.base import MessageBase

import os
import asyncio
import zmq.asyncio


PORT = 9432
PORT_INSECURE = 9433
PROTOCOL = 'tcp'

PUB_SOCK_KEY = 'default_pub'
SUB_SOCK_KEY = 'default_sub'
DEFAULT_DOMAIN = '*'


def wrap_func(fn, *args, **kwargs):
    def wrapper():
        return fn(*args, **kwargs)
    return wrapper


class RouterAuthTester(Worker):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.router = None

    def start(self):
        self.loop.run_forever()

    def create_router_socket(self, identity: bytes, secure=True, domain=DEFAULT_DOMAIN, name='Router', debug=True):
        self.router = self.manager.create_socket(zmq.ROUTER, secure=secure, domain=domain, name=name)
        self.router.setsockopt(zmq.IDENTITY, identity)
        self.router.add_handler(handler_func=self.handle_router_msg, start_listening=True)
        if debug:
            self.router.setsockopt(zmq.ROUTER_MANDATORY, 1)

    def connect_router_socket(self, vk='', ip='', port=PORT, protocol=PROTOCOL):
        assert self.router, "create_router_socket must be called first"
        self.router.connect(ip=ip, vk=vk, port=port, protocol=protocol)

    def bind_router_socket(self, ip, port=PORT, protocol=PROTOCOL):
        assert self.router, "create_router_socket must be called first"
        self.router.bind(ip=ip, port=port, protocol=protocol)

    def send_raw_msg(self, id_frame: bytes, msg: bytes):
        self.router.send_multipart([id_frame, msg])

    def send_msg(self, msg: MessageBase, header: bytes):
        self.router.send_msg(msg, header)

    def handle_router_msg(self, frames):
        self.log.important("Router got msg with frames {}".format(frames))



from contracting.client import ContractingClient
from contracting.db.driver import ContractDriver
from cilantro_ee.services.storage.vkbook import PhoneBook
import capnp
import os
from cilantro_ee.messages import capnp as schemas

blockdata_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/blockdata.capnp')

PENDING_REWARDS_KEY = '__rewards'


class RewardManager:
    def __init__(self, driver=ContractDriver(), client=ContractingClient()):
        self.driver = driver
        self.client = client

        self.stamp_contract = self.client.get_contract('stamp_cost')
        self.reward_contract = self.client.get_contract('rewards')
        self.currency_contract = self.client.get_contract('currency')

        assert self.stamp_contract is not None, 'Stamp contract not in state.'
        assert self.reward_contract is not None, 'Reward contract not in state.'
        assert self.currency_contract is not None, 'Currency contract not in state.'

    def issue_rewards(self):
        master_ratio, delegate_ratio, burn_ratio, foundation_ratio = self.reward_ratio
        pending_rewards = self.get_pending_rewards()

        masters = PhoneBook.masternodes
        delegates = PhoneBook.delegates

        master_reward = (master_ratio * pending_rewards) / len(masters)
        delegate_reward = (delegate_ratio * pending_rewards) / len(delegates)
        foundation_reward = foundation_ratio * pending_rewards

        for m in masters:
            self.add_to_balance(vk=m, amount=master_reward)

        for d in delegates:
            self.add_to_balance(vk=d, amount=delegate_reward)

        self.set_pending_rewards(0)

    def add_to_balance(self, vk, amount):
        current_balance = self.currency_contract.quick_read(variable='balances', key=vk) or 0
        self.currency_contract.quick_write(variable='balances', key=vk, value=amount + current_balance)

    def get_pending_rewards(self):
        return self.driver.get(PENDING_REWARDS_KEY) or 0

    def set_pending_rewards(self, value):
        self.driver.set(PENDING_REWARDS_KEY, value=value)

    @property
    def stamps_per_tau(self):
        return self.stamp_contract.quick_read('S', 'rate')

    @staticmethod
    def stamps_in_block(block):
        total = 0

        for sb in block.subBlocks:
            for tx in sb.transactions:
                total += tx.stampsUsed

        return total

    @staticmethod
    def stamps_in_subblock(subblock):
        total = 0

        for tx in subblock.transactions:
            total += tx.stampsUsed

        return total

    def add_pending_rewards(self, subblock):
        current_rewards = self.get_pending_rewards()
        used_stamps = self.stamps_in_subblock(subblock)

        rewards_as_tau = used_stamps / self.stamps_per_tau
        self.set_pending_rewards(current_rewards + rewards_as_tau)

    @property
    def reward_ratio(self):
        return self.reward_contract.quick_read(variable='value')



from cilantro_ee.utils.lprocess import LProcess
from cilantro_ee.utils.utils import is_valid_hex, int_to_bytes
from cilantro_ee.utils.lazy_property import lazy_property, set_lazy_property, lazy_func
from cilantro_ee.utils.hasher import Hasher



from unittest import TestCase
from cilantro_ee.services.storage.state import MetaDataStorage
from cilantro_ee.messages.transaction.data import TransactionDataBuilder, TransactionData
from cilantro_ee.messages.block_data.sub_block import SubBlock, SubBlockBuilder
from cilantro_ee.messages.block_data.block_data import GENESIS_BLOCK_HASH, BlockData
import json

class TestMetaDataStorage(TestCase):
    def setUp(self):
        self.db = MetaDataStorage()
        self.db.flush()

    def tearDown(self):
        self.db.flush()

    def test_init(self):
        self.assertIsNotNone(self.db)

    def test_set_latest_block_hash_not_64_chars(self):
        bhash = b'a' * 6
        with self.assertRaises(AssertionError):
            self.db.set_latest_block_hash(bhash)

    def test_set_latest_block_hash_not_hex_fails(self):
        bhash = 'x' * 32
        with self.assertRaises(ValueError):
            self.db.set_latest_block_hash(bhash)

    def test_set_latest_block_hash_returns_when_successful(self):
        bhash = b'a' * 32

        self.db.set_latest_block_hash(bhash)

    def test_get_latest_block_hash_none(self):
        expected = b'\00' * 32

        got = self.db.get_latest_block_hash()

        self.assertEqual(expected, got)

    def test_get_latest_block_hash_after_setting(self):
        expected = b'a' * 32

        self.db.set_latest_block_hash(expected)

        got = self.db.get_latest_block_hash()

        self.assertEqual(expected, got)

    def test_latest_block_hash_as_property(self):
        expected = b'a' * 32

        self.db.latest_block_hash = expected

        got = self.db.latest_block_hash

        self.assertEqual(expected, got)

    def test_set_latest_block_num_not_number(self):
        num = 'a'
        with self.assertRaises(ValueError):
            self.db.set_latest_block_num(num)

    def test_set_latest_block_num_negative_fails(self):
        num = -1000
        with self.assertRaises(AssertionError):
            self.db.set_latest_block_num(num)

    def test_set_latest_block_num_returns_when_successful(self):
        num = 64

        self.db.set_latest_block_num(num)

    def test_get_latest_block_num_none(self):
        got = self.db.get_latest_block_num()

        self.assertEqual(0, got)

    def test_get_latest_block_num_after_setting(self):
        num = 64

        self.db.set_latest_block_num(num)

        got = self.db.get_latest_block_num()

        self.assertEqual(num, got)

    def test_get_latest_block_num_as_property(self):
        num = 64

        self.db.latest_block_num = num

        got = self.db.latest_block_num

        self.assertEqual(num, got)



"""
Utility functions for tests.
"""
import random
import hashlib
from struct import pack

from zmq.utils.z85 import decode, encode
from nacl.public import PrivateKey
from nacl.signing import SigningKey, VerifyKey
from nacl.bindings import crypto_sign_ed25519_sk_to_curve25519

from cilantro_ee.services.overlay.kademlia.node import Node

def mknode(node_id=None, ip=None, port=None, intid=None):
    """
    Make a node.  Created a random id if not specified.
    """
    if intid is not None:
        node_id = pack('>l', intid)
    if not node_id:
        randbits = str(random.getrandbits(255))
        node_id = hashlib.sha1(randbits.encode()).digest()
    return Node(node_id, ip, port)

def genkeys(sk_hex):
    sk = SigningKey(seed=bytes.fromhex(sk_hex))
    vk = sk.verify_key.encode().hex()
    public_key = VerifyKey(bytes.fromhex(vk)).to_curve25519_public_key()._public_key
    private_key = crypto_sign_ed25519_sk_to_curve25519(sk._signing_key)
    return {
        'sk': sk_hex,
        'vk': vk,
        'public_key': public_key.hex(),
        'secret_key': private_key.hex(),
        'private_key': encode(private_key),
        'curve_key': encode(public_key)
    }

class FakeProtocol:
    def __init__(self, sourceID, ksize=20):
        self.router = RoutingTable(self, ksize, Node(sourceID))
        self.storage = {}
        self.sourceID = sourceID



import zmq
import zmq.asyncio
import asyncio
import os
import uuid
import json
from cilantro_ee.constants.overlay_network import EVENT_URL, CMD_URL
from cilantro_ee.core.logger.base import get_logger

log = get_logger('Overlay.Client')
# Sends the following multipart message
# [Function name encoded and event ID encoded], [Args encoded if it can be encoded], [KWards encoded if they can be]
# Returns event ID.


def command(fn):
    def _command(self, *args, **kwargs):
        log.info('COMMAND RUN: {}'.format(fn.__name__.encode()))
        event_id = uuid.uuid4().hex
        self.cmd_sock.send_multipart(
            [fn.__name__.encode(), event_id.encode()] + \
            [arg.encode() if hasattr(arg, 'encode') else str(arg).encode() for arg in args] + \
            [kwargs[k].encode() if hasattr(kwargs[k], 'encode') else str(kwargs[k]).encode() for k in kwargs])
        return event_id

    return _command


# Test Overlay Client. Try publishing mock messages to the event socket

class OverlayClient:
    def __init__(self, reply_handler, event_handler, ctx, name=None):
        self.name = name or str(os.getpid())

        self.log = get_logger('Overlay.Client.{}'.format(name))
        self.loop = asyncio.get_event_loop()
        self.ctx = ctx

        self.cmd_sock = self.ctx.socket(socket_type=zmq.DEALER)
        self.cmd_sock.setsockopt(zmq.IDENTITY, self.name.encode())
        self.cmd_sock.connect(CMD_URL)

        self.evt_sock = self.ctx.socket(socket_type=zmq.SUB)
        self.evt_sock.setsockopt(zmq.SUBSCRIBE, b"")
        self.evt_sock.connect('tcp://127.0.0.1:10003')

        self.tasks = [
            self.reply_listener(reply_handler),
            self.event_listener(event_handler),
        ]

        self.running = False

    @command
    def ready(self, *args, **kwargs):
        # b'ready'
        pass

    def get_ip_from_vk(self, *args, **kwargs):
        # b'get_ip_from_vk
        event_id = uuid.uuid4().hex
        self.cmd_sock.send_multipart(
            [b'get_ip_from_vk', event_id.encode()] + \
            [arg.encode() if hasattr(arg, 'encode') else str(arg).encode() for arg in args] + \
            [kwargs[k].encode() if hasattr(kwargs[k], 'encode') else str(kwargs[k]).encode() for k in kwargs]
        )
        return event_id

    async def event_listener(self, event_handler):
        self.log.success('Listening for overlay events.')
        while self.running:
            msg = await self.evt_sock.recv()
            self.log.success("OverlayClient received event {}".format(msg))

            response = json.loads(msg.decode())

            if isinstance(response, dict):
                event_handler(response)

    async def reply_listener(self, reply_handler):
        self.log.success("Listening for overlay replies over {}".format(CMD_URL))
        while self.running:
            msg = await self.cmd_sock.recv_multipart()
            event = json.loads(msg[-1])
            self.log.success("OverlayClient received reply {} and event {}".format(msg, event))
            reply_handler(event)

    def teardown(self):
        self.cmd_sock.close()
        self.evt_sock.close()

    async def start(self):
        self.running = True
        asyncio.ensure_future(asyncio.gather(*self.tasks))
        log.info('Overlay Client started!')

    def stop(self):
        self.running = False



import asyncio, time
from threading import Timer

def async_run_for(delay):
    def _run(fn, *arg, **kwargs):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        def _stop():
            loop.call_soon_threadsafe(loop.stop)
        def _fn(self):
            stopper = Timer(delay, _stop)
            stopper.start()
            fn(self)
            loop.run_forever()
        return _fn
    return _run



from cilantro_ee.services.storage.vkbook import PhoneBook
import math


POW_COMPLEXITY = ''  # More '0's means more complicated POWs. Empty string basically disables POW

BLOCK_HEART_BEAT_INTERVAL = 60
INPUT_BAG_TIMEOUT = 10

# In reality, these should be inferred from VKBook instead of hard-coded, once we start using smart contracts for
# some of these config constants

NUM_MASTERS = PhoneBook.num_boot_masternodes
NUM_WITNESSES = len(PhoneBook.witnesses)
NUM_DELEGATES = PhoneBook.num_boot_delegates
NUM_NODES = NUM_MASTERS + NUM_WITNESSES + NUM_DELEGATES

# How long each Node will wait for the rest of the network to come online before an error is raised
MAX_BOOT_WAIT = 600


# ///////////////////////////////////////////////
# Consensus
# ///////////////////////////////////////////////
DELEGATE_MAJORITY = math.ceil(NUM_DELEGATES * 2 / 3)
MASTERNODE_MAJORITY = math.ceil(NUM_MASTERS * 2 / 3)

BLOCK_NOTIFICATION_QUORUM = (NUM_MASTERS + 1) // 2
FAILED_BLOCK_NOTIFICATION_QUORUM = NUM_MASTERS - BLOCK_NOTIFICATION_QUORUM + 1


# ///////////////////////////////////////////////
# Block and Sub-block
# ///////////////////////////////////////////////

DUMP_TO_CACHE_EVERY_N_BLOCKS = 5
# A Masternode expects to produce a block or empty block every BLOCK_TIMEOUT seconds or he will send a SkipBlockNotif
BLOCK_PRODUCTION_TIMEOUT = 60
TRANSACTIONS_PER_SUB_BLOCK = 20

_MAX_SUB_BLOCK_BUILDERS = 4
_MIN_BLOCKS = 1  
NUM_SUB_BLOCKS = NUM_MASTERS
NUM_BLOCKS = max(_MIN_BLOCKS, (NUM_SUB_BLOCKS + _MAX_SUB_BLOCK_BUILDERS - 1) // _MAX_SUB_BLOCK_BUILDERS) 

NUM_SB_PER_BLOCK = (NUM_SUB_BLOCKS + NUM_BLOCKS - 1) // NUM_BLOCKS

NUM_SB_BUILDERS = min(_MAX_SUB_BLOCK_BUILDERS, NUM_SB_PER_BLOCK)
NUM_SB_PER_BUILDER = (NUM_SUB_BLOCKS + NUM_SB_BUILDERS - 1) // NUM_SB_BUILDERS
NUM_SB_PER_BLOCK_PER_BUILDER = (NUM_SB_PER_BLOCK + NUM_SB_BUILDERS - 1) // NUM_SB_BUILDERS

assert NUM_SUB_BLOCKS >= NUM_BLOCKS, "num_blocks {} cannot be more than num_sub_blocks {}".format(NUM_BLOCKS, NUM_SUB_BLOCKS)
assert NUM_SUB_BLOCKS/NUM_SB_PER_BLOCK == NUM_BLOCKS, "NUM_SUB_BLOCKS/NUM_SB_PER_BLOCK should equal NUM_BLOCKS.\n" \
                                                      "{} / {} != {}".format(NUM_SUB_BLOCKS, NUM_SB_PER_BLOCK, NUM_BLOCKS)
assert NUM_SB_PER_BLOCK_PER_BUILDER >= 1, "num_sub_blocks_per_block_per_builder {} cannot less than 1".format(NUM_SB_PER_BLOCK_PER_BUILDER)


# ///////////////////////////////////////////////
# Transaction Batcher
# ///////////////////////////////////////////////
BATCH_SLEEP_INTERVAL = 1


# ///////////////////////////////////////////////
# Delegate
# ///////////////////////////////////////////////


# ///////////////////////////////////////////////
# Seneca Interpreter
# ///////////////////////////////////////////////
DECIMAL_PRECISION = 18


# ///////////////////////////////////////////////
# Test Flags
# ///////////////////////////////////////////////
SHOULD_MINT_WALLET = True
# NUM_WALLETS_TO_MINT = 50

NUM_WALLETS_TO_MINT = 100
MINT_AMOUNT = 10 ** 10



from cilantro_ee.core.logger import get_logger
from cilantro_ee.utils.test.mp_test_case import MPTestCase
from cilantro_ee.utils.test.mp_test import MPTesterBase, mp_testable
from cilantro_ee.utils.test.god import God
from unittest.mock import MagicMock
from cilantro_ee.protocol.transport.composer import Composer
from cilantro_ee.protocol.executors.manager import ExecutorManager
from cilantro_ee.protocol.states.statemachine import StateMachine
from cilantro_ee.nodes.masternode.masternode import Masternode
from cilantro_ee.nodes.delegate.delegate import Delegate
from cilantro_ee.nodes.witness.witness import Witness
from cilantro_ee.nodes.base import NodeBase
from cilantro_ee.services.overlay.daemon import OverlayServer
from cilantro_ee.utils.lprocess import LProcess
# from cilantro_ee.services.storage.db import DB
from cilantro_ee.utils.test.pubsub_auth import PubSubAuthTester
from cilantro_ee.utils.test.router_auth import RouterAuthTester
import asyncio
import zmq.asyncio
import os
from cilantro_ee.services.storage.vkbook import VKBook


@mp_testable(NodeBase)
class MPNodeBase(MPTesterBase):
    @classmethod
    def build_obj(cls, sk, name='Node'):
        VKBook.setup()
        ip = os.getenv('HOST_IP', '127.0.0.1')
        loop = asyncio.get_event_loop()
        asyncio.set_event_loop(loop)
        obj = NodeBase(ip=ip, signing_key=sk, loop=loop, name=name)

        return obj, loop, []


@mp_testable(PubSubAuthTester)
class MPPubSubAuth(MPTesterBase):
    @classmethod
    def build_obj(cls, sk, name='') -> tuple:
        VKBook.setup()
        # # DEBUG -- TODO DELETE
        # print("VKBook on REMOTE MACHINE")
        # from cilantro_ee.services.storage.vkbook import VKBook
        # VKBook.test_print_nodes()
        # # END DEBUG
        loop = asyncio.get_event_loop()
        asyncio.set_event_loop(loop)

        # Start Overlay Process
        overlay_proc = LProcess(target=OverlayServer, args=(sk,))
        overlay_proc.start()

        obj = PubSubAuthTester(sk, name=name, loop=loop)

        return obj, loop, []


@mp_testable(RouterAuthTester)
class MPRouterAuth(MPTesterBase):
    @classmethod
    def build_obj(cls, sk, name='') -> tuple:
        VKBook.setup()
        loop = asyncio.get_event_loop()
        asyncio.set_event_loop(loop)

        # Start Overlay Process
        overlay_proc = LProcess(target=OverlayServer, args=(sk,))
        overlay_proc.start()

        obj = RouterAuthTester(sk, name=name, loop=loop)

        return obj, loop, []


@mp_testable(Composer)
class MPComposer(MPTesterBase):
    @classmethod
    def build_obj(cls, sk, name='') -> tuple:
        VKBook.setup()
        loop = asyncio.get_event_loop()
        asyncio.set_event_loop(loop)
        ctx = zmq.asyncio.Context()

        # Start Overlay Process
        overlay_proc = LProcess(target=OverlayServer, args=(sk,))
        overlay_proc.start()

        router = MagicMock()
        ip = os.getenv('HOST_IP', '127.0.0.1')
        manager = ExecutorManager(signing_key=sk, router=router, name=name, loop=loop, context=ctx)
        composer = Composer(manager=manager, signing_key=sk, ip=ip, name=name)

        return composer, loop, []


@mp_testable(StateMachine)
class MPStateMachine(MPTesterBase):
    @classmethod
    def build_obj(cls, sm_class, name=''):
        VKBook.setup()
        # These 2 lines are probs unnecessary
        loop = asyncio.get_event_loop()
        asyncio.set_event_loop(loop)

        sm = sm_class()
        assert isinstance(sm, StateMachine), "Class {} did not return a StateMachine instance once called".format(sm_class)

        # NOTE -- the testing framework should start the loop, so we shouldn't have to worry bout that here
        return sm, loop, []


@mp_testable(God)
class MPGod(MPTesterBase):
    @classmethod
    def build_obj(cls) -> tuple:
        VKBook.setup()
        loop = asyncio.get_event_loop()
        god = God(loop=loop)

        return god, loop, [god.interface._recv_messages()]


# def _build_node(signing_key, name='', node_cls=None) -> tuple:
#     assert node_cls and name, "This is an abstract class. Subclasses must pass in node_cls and name."
#     assert node_cls in (Witness, Masternode, Delegate), "node_cls must be Witness/Masternode/Delegate, not {}".format(node_cls)
#
#     loop = asyncio.get_event_loop()
#     asyncio.set_event_loop(loop)
#
#     with DB(should_reset=True) as db:
#         pass
#
#     ip = os.getenv('HOST_IP', '127.0.0.1')
#
#     node = NodeFactory._build_node(loop=loop, signing_key=signing_key, ip=ip, node_cls=node_cls, name=name)
#     node.start(start_loop=False)
#
#     tasks = node.tasks + [node.composer.interface._recv_messages()]
#
#     return node, loop, tasks
#
#
# @mp_testable(Masternode)
# class MPMasternode(MPTesterBase):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#
#         # Set God's Masternode URL to use this guy updated port if we are running on VM
#         if MPTestCase.vmnet_test_active:
#             node_ports = MPTestCase.ports[self.container_name]
#             assert '8080' in node_ports, "Expected port 8080 to be available in Falcon's docker external port thing"
#             God.set_mn_url(ip='127.0.0.1', port=node_ports['8080'].split(':')[-1])
#
#     @classmethod
#     def build_obj(cls, signing_key, name='Masternode') -> tuple:
#         return _build_node(signing_key=signing_key, name=name, node_cls=Masternode)
#
#
# @mp_testable(Witness)
# class MPWitness(MPTesterBase):
#     @classmethod
#     def build_obj(cls, signing_key, name='Witness') -> tuple:
#         return _build_node(signing_key=signing_key, name=name, node_cls=Witness)
#
#
# @mp_testable(Delegate)
# class MPDelegate(MPTesterBase):
#     @classmethod
#     def build_obj(cls, signing_key, name='Delegate') -> tuple:
#         return _build_node(signing_key=signing_key, name=name, node_cls=Delegate)



from cilantro_ee.nodes.masternode.masternode import Masternode
from cilantro_ee.nodes.delegate.delegate import Delegate
from cilantro_ee.constants import conf

import time
from pymongo import MongoClient

MASTERNODE = 0
DELEGATE = 1
WITNESS = 2


def wait_for_redis():
    time.sleep(2)


def wait_for_mongo():
    while True:
        try:
            info = MongoClient().server_info()
            print("Mongo ready! Server info:\n{}".format(info))
            break
        except:
            print("Waiting for Mongo to be ready...")
            time.sleep(1)


def start_node(signing_key, node_type):
    wait_for_redis()

    if node_type == MASTERNODE:
        wait_for_mongo()

        Masternode(ip=conf.HOST_IP, name='Masternode', signing_key=signing_key)

    elif node_type == DELEGATE:
        Delegate(ip=conf.HOST_IP, name='Delegate', signing_key=signing_key)



from unittest import TestCase
from cilantro_ee.core.crypto import wallet
from cilantro_ee.messages.consensus.merkle_signature import build_test_merkle_sig, MerkleSignature
import json


class TestMerkleSignature(TestCase):

    def test_valid_creation(self):
        """
        Tests that a MerkleSignature created with some argument has the expected properties
        """
        msg = b'this is a pretend merkle tree hash'
        timestamp = 'now'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)
        ms = MerkleSignature.create(sig_hex=signature, timestamp=timestamp, sender=vk)

        self.assertEqual(ms.signature, signature)
        self.assertEqual(ms.timestamp, timestamp)
        self.assertEqual(ms.sender, vk)

    def test_invalid_signature_nonhex(self):
        """
        Tests that an error is raised during validation if an invalid signature is passed (nonhex, or length != 128)
        """
        msg = b'this is a pretend merkle tree hash'
        timestamp = 'now'
        sk, vk = wallet.new()

        # Test nonhex signature (but valid length)
        sig = ''.join(('X' for _ in range(128)))
        nonhex = MerkleSignature.create(sig_hex=sig, timestamp=timestamp, sender=vk, validate=False)
        self.assertRaises(Exception, nonhex.validate)

    def test_invalid_signature_bad_length(self):
        # Test signature incorrect length (but valid hex)
        sk, vk = wallet.new()
        timestamp = 'now'
        sig = ''.join(('A' for _ in range(100)))
        wrong_len = MerkleSignature.create(sig_hex=sig, timestamp=timestamp, sender=vk, validate=False)
        self.assertRaises(Exception, wrong_len.validate)

    def test_validate_catches_bad_sig(self):
        # test that validate is called by default and throws an exception with bad sig
        sk, vk = wallet.new()
        sig = ''.join(('X' for _ in range(128)))
        timestamp = 'now'

        self.assertRaises(Exception, MerkleSignature.create, sig_hex=sig, timestamp=timestamp, sender=vk)

    def test_invalid_sender_wrong_sender(self):
        """
        Tests that an error is raised during creation if an invalid sender field is passed in. A sender should be a
        64 character hex string verifying key.
        """
        # Test an error is thrown when MerkleSignature created with a sender that is not the correct public key
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)

        timestamp = 'now'
        vk_bad = wallet.new()[1]  # different verifying (public) key
        bad_public_key = MerkleSignature.create(sig_hex=signature, timestamp=timestamp, sender=vk_bad)
        self.assertRaises(Exception, bad_public_key)

    def test_valid_sender(self):
        # Confirm no error when correct public key is used
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)

        timestamp = 'now'
        MerkleSignature.create(sig_hex=signature, timestamp=timestamp, sender=vk)  # no error thrown

    def test_invalid_sender_bad_hash(self):
        # Test an error is thrown when created with a sender of not valid hash
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)

        timestamp = 'now'
        vk_bad_hash = ''.join('Z' for _ in range(64))  # verifying (public) key with bad hash
        self.assertRaises(Exception, MerkleSignature.create(sig_hex=signature, timestamp=timestamp, sender=vk_bad_hash,
                                                            validate=False))

    def test_invalid_sender_bad_pk_length(self):
        # Test an error is thrown when created with a sender of invalid length (not 64)
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)

        timestamp = 'now'
        vk_bad_length = ''.join('e' for _ in range(75))  # verifying (public) key with bad length

        self.assertRaises(Exception, MerkleSignature.create(sig_hex=signature, timestamp=timestamp,
                                                            sender=vk_bad_length, validate=False))

    def test_invalid_timestamp(self):
        """
        Test that if the timestamp field is not formatted as expected an error will be thrown
        """
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)

        timestamp = 99
        self.assertRaises(TypeError, MerkleSignature.create(sig_hex=signature, timestamp=timestamp, sender=vk))

    def test_deserialization_invalid_json(self):
        """
        Tests that attempting to deserialize a MerkleSignature from a bad binary (not json, not valid fields, ect) throws
        as error
        """
        # Test bad (invalid) JSON
        bad_json = b'lololololol()**[][]XXX-----!!!!!'
        self.assertRaises(json.decoder.JSONDecodeError, MerkleSignature.from_bytes, bad_json)

    def test_deserialization_missing_fields(self):
        # Test valid json but missing fields
        sig = ''.join(('A' for _ in range(128)))
        d = {MerkleSignature.SIG: sig, MerkleSignature.TS: 'now'}
        binary = json.dumps(d).encode()
        self.assertRaises(AssertionError, MerkleSignature.from_bytes, binary)

    def test_deserialization_bad_sk_length(self):
        # Test valid json but signature (private key) is of the wrong length
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        sig = ''.join(('A' for _ in range(100)))

        d = {MerkleSignature.SIG: sig, MerkleSignature.TS: 'now', MerkleSignature.SENDER: vk}
        binary = json.dumps(d).encode()
        self.assertRaises(Exception, MerkleSignature.from_bytes, binary)

    def test_deserialization_bad_sk_hex(self):
        # Test valid json but signature (private key) not proper hex
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        sig = ''.join(('Z' for _ in range(128)))

        d = {MerkleSignature.SIG: sig, MerkleSignature.TS: 'now', MerkleSignature.SENDER: vk}
        binary = json.dumps(d).encode()
        self.assertRaises(Exception, MerkleSignature.from_bytes, binary)

    def test_deserialization_valid_json(self):
        # Test valid json throws no errors
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)

        d = {MerkleSignature.SIG: signature, MerkleSignature.TS: 'now', MerkleSignature.SENDER: vk}
        binary = json.dumps(d).encode()
        MerkleSignature.from_bytes(binary)

    def test_serialization(self):
        """
        Tests that a created block data reply successfully serializes and deserializes. The deserialized object should
        have the same properties as the original one before it was serialized.
        """
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)

        timestamp = 'now'
        valid_merkle_sig = MerkleSignature.create(sig_hex=signature, timestamp=timestamp, sender=vk)

        valid_merkle_sig_serialized = valid_merkle_sig.serialize()

        clone = MerkleSignature.from_bytes(valid_merkle_sig_serialized)

        self.assertEqual(valid_merkle_sig.signature, clone.signature)
        self.assertEqual(valid_merkle_sig.timestamp, clone.timestamp)
        self.assertEqual(valid_merkle_sig.sender, clone.sender)

    def test_verify_valid_ms(self):
        """
        Tests that MerkleSignature.verify(...) returns true given a proper msg and vk
        """
        # Test merkle tree verify() validates correct verifying (public) keys
        msg = b'this is a pretend merkle tree hash'
        timestamp = 'now'
        sk, vk = wallet.new()
        signature = wallet.sign(sk, msg)
        ms = MerkleSignature.create(sig_hex=signature, timestamp=timestamp, sender=vk)

        self.assertTrue(ms.verify(msg))

    def test_build_test_merkle_sig(self):
        """
        Tests build_test_contender. This is used exclusively unit tests, so we basically just want to make sure it
        doesn't blow up here first.
        """
        sig = build_test_merkle_sig()

    def test_create_from_payload(self):
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        sig = MerkleSignature.create_from_payload(signing_key=sk, payload=msg)
        self.assertTrue(sig.verify(msg))

    def test_create_from_payload_cloned(self):
        msg = b'this is a pretend merkle tree hash'
        sk, vk = wallet.new()
        sig = MerkleSignature.create_from_payload(signing_key=sk, payload=msg)
        clone = MerkleSignature.from_bytes(sig.serialize())
        self.assertTrue(clone.verify(msg))



from sanic import Sanic
from sanic import response
#from secure import SecureHeaders
from cilantro_ee.logger.base import get_logger
from sanic_cors import CORS
import json as _json
from contracting.client import ContractingClient

from cilantro_ee.constants import conf
from cilantro_ee.utils.hasher import Hasher
from cilantro_ee.core.messages.capnp_impl.capnp_impl import pack
from cilantro_ee.services.storage.master import MasterStorage
from cilantro_ee.services.storage.state import MetaDataStorage
from cilantro_ee.core.nonces import NonceManager

from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message

from multiprocessing import Queue
import ast
import ssl
import time

import hashlib

WEB_SERVER_PORT = 8080
SSL_WEB_SERVER_PORT = 443
NUM_WORKERS = 2

app = Sanic(__name__)

ssl_enabled = False
ssl_cert = '~/.ssh/server.csr'
ssl_key = '~/.ssh/server.key'

CORS(app, automatic_options=True)
log = get_logger("MN-WebServer")
client = ContractingClient()
metadata_driver = MetaDataStorage()
nonce_manager = NonceManager()

static_headers = {}


# @app.middleware('response')
# async def set_secure_header(request, response):
#     SecureHeaders.sanic(response)

@app.route("/")
async def hello(request):
    return response.text("hello world")

# ping to check whether server is online or not
@app.route("/ping", methods=["GET","OPTIONS",])
async def ping(request):
    return response.json({'Hello', True})


@app.route('/id', methods=['GET'])
async def get_id(request):
    return response.json({'verifying_key': conf.HOST_VK})


@app.route('/nonce/<vk>', methods=['GET'])
async def get_nonce(request, vk):
    # Might have to change this sucker from hex to bytes.
    pending_nonce = nonce_manager.get_pending_nonce(processor=conf.HOST_VK, sender=bytes.fromhex(vk))

    log.info('Pending nonce: {}'.format(pending_nonce))

    if pending_nonce is None:
        nonce = nonce_manager.get_nonce(processor=conf.HOST_VK, sender=bytes.fromhex(vk))
        log.info('Pending nonce was none so got nonce which is {}'.format(nonce))
        if nonce is None:
            pending_nonce = 0
            log.info('Nonce was now so pending nonce is now zero.')
        else:
            pending_nonce = nonce
            log.info('Nonce was not none so setting pending nonce to it.')

    return response.json({'nonce': pending_nonce, 'processor': conf.HOST_VK.hex(), 'sender': vk})


@app.route('/epoch', methods=['GET'])
async def get_epoch(request):
    epoch_hash = metadata_driver.latest_epoch_hash
    block_num = metadata_driver.latest_block_num

    e = (block_num // conf.EPOCH_INTERVAL) + 1
    blocks_until_next_epoch = (e * conf.EPOCH_INTERVAL) - block_num

    return response.json({'epoch_hash': epoch_hash.hex(),
                 'blocks_until_next_epoch': blocks_until_next_epoch})


@app.route("/", methods=["POST","OPTIONS",])
async def submit_transaction(request):
    if app.queue.full():
        return response.json({'error': "Queue full. Resubmit shortly."}, status=503)

    # Try to deserialize transaction.
    try:
        tx_bytes = request.body
        tx = Message.unpack_message(pack(int(MessageType.TRANSACTION)), tx_bytes)

    except Exception as e:
        return response.json({'error': 'Malformed transaction.'.format(e)}, status=400)

    # Try to put it in the request queue.
    try:
        app.queue.put_nowait(tx)
    except:
        return response.json({'error': "Queue full. Resubmit shortly."}, status=503)

    h = hashlib.sha3_256()
    h.update(tx_bytes)
    tx_hash = h.digest()

    return response.json({'success': 'Transaction successfully submitted to the network.',
                 'hash': tx_hash.hex()})


# Returns {'contracts': JSON List of strings}
@app.route('/contracts', methods=['GET'])
async def get_contracts(request):
    contracts = client.get_contracts()
    return response.json({'contracts': contracts})


@app.route('/contracts/<contract>', methods=['GET'])
async def get_contract(request, contract):
    contract_code = client.raw_driver.get_contract(contract)

    if contract_code is None:
        return response.json({'error': '{} does not exist'.format(contract)}, status=404)
    return response.json({'name': contract, 'code': contract_code}, status=200)


@app.route("/contracts/<contract>/methods", methods=['GET'])
async def get_methods(request, contract):
    contract_code = client.raw_driver.get_contract(contract)

    if contract_code is None:
        return response.json({'error': '{} does not exist'.format(contract)}, status=404)

    tree = ast.parse(contract_code)

    function_defs = [n for n in ast.walk(tree) if isinstance(n, ast.FunctionDef)]

    funcs = []
    for definition in function_defs:
        func_name = definition.name
        kwargs = [arg.arg for arg in definition.args.args]

        funcs.append({'name': func_name, 'arguments': kwargs})

    return response.json({'methods': funcs}, status=200)


@app.route('/contracts/<contract>/<variable>')
async def get_variable(request, contract, variable):
    contract_code = client.raw_driver.get_contract(contract)

    if contract_code is None:
        return response.json({'error': '{} does not exist'.format(contract)}, status=404)

    key = request.args.get('key')

    k = client.raw_driver.make_key(key=contract, field=variable, args=key)
    response = client.raw_driver.get(k)

    if response is None:
        return response.json({'value': None}, status=404)
    else:
        return response.json({'value': response}, status=200)


# Expects json object such that:
'''
{
    'name': 'string',
    'code': 'string'
}
'''

@app.route("/latest_block", methods=["GET","OPTIONS",])
async def get_latest_block(request):
    index = MasterStorage.get_last_n(1)
    latest_block_hash = index.get('blockHash')
    return response.json({ 'hash': '{}'.format(latest_block_hash) })


@app.route('/blocks', methods=["GET","OPTIONS",])
async def get_block(request):
    if 'number' in request.json:
        num = request.json['number']
        block = MasterStorage.get_block(num)
        if block is None:
            return response.json({'error': 'Block at number {} does not exist.'.format(num)}, status=400)
    # TODO check block by hash isn't implemented
    # else:
    #     _hash = request.json['hash']
    #     block = MasterStorage.get_block(hash)
    #     if block is None:
    #         return _respond_to_request({'error': 'Block with hash {} does not exist.'.format(_hash)}, 400)

    return response.json(_json.dumps(block))


def start_webserver(q):
    app.queue = q
    if ssl_enabled:
        context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        context.load_cert_chain(ssl_cert, keyfile=ssl_key)
        app.run(host='0.0.0.0', port=SSL_WEB_SERVER_PORT, workers=NUM_WORKERS, debug=False, access_log=False, ssl=context)
    else:
        app.run(host='0.0.0.0', port=WEB_SERVER_PORT, workers=NUM_WORKERS, debug=False, access_log=False)


if __name__ == '__main__':
    import pyximport; pyximport.install()
    if not app.config.REQUEST_MAX_SIZE:
        app.config.update({
            'REQUEST_MAX_SIZE': 10000,
            'REQUEST_TIMEOUT': 20
        })
    start_webserver(Queue())



from cilantro_ee.messages import capnp as schemas
import os
import capnp

envelope_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/envelope.capnp')

e = envelope_capnp.Message.new_message()

e.payload = b'poo'

print(e.to_bytes_packed())



from unittest import TestCase
from cilantro_ee.services.storage.master import DistributedMasterStorage
from cilantro_ee.core.crypto import wallet
from cilantro_ee.services.storage.vkbook import PhoneBook, VKBook


class TestDistributedMasterStorage(TestCase):
    def setUp(self):
        sk, vk = wallet.new()
        self.db = DistributedMasterStorage(key=sk)

    def tearDown(self):
        self.db.drop_collections()

    def test_init(self):
        self.assertTrue(self.db)

    def test_get_masterset_test_hook_false(self):
        self.db.test_hook = False

        self.assertEqual(len(PhoneBook.masternodes), self.db.get_master_set())

    def test_get_masterset_test_hook_true(self):
        am = self.db.active_masters

        self.db.test_hook = True

        self.assertEqual(am, self.db.get_master_set())

    def test_set_mn_id_test_hook_true(self):
        mn_id = self.db.mn_id

        self.db.test_hook = True

        self.assertEqual(mn_id, self.db.set_mn_id(vk='0'))

    def test_set_mn_id_test_hook_false_master_not_in_active_masters(self):
        vk = 'IMPOSSIBLE WALLET'

        self.db.test_hook = False

        success = self.db.set_mn_id(vk)

        self.assertEqual(self.db.mn_id, -1)
        self.assertFalse(success)

    def test_set_mn_id_test_hook_false_master_in_active_masters(self):
        vk = PhoneBook.masternodes[0]

        self.db.test_hook = False

        success = self.db.set_mn_id(vk)

        self.assertEqual(self.db.mn_id, 0)
        self.assertTrue(success)

    def test_rep_pool_size_fails_when_active_masters_less_than_rep_factor(self):
        self.db.rep_factor = 999
        self.assertEqual(self.db.rep_pool_sz(), -1)

    def test_rep_pool_size_returns_correctly_rounded_pool_size_when_enough_masters_present(self):
        self.db.test_hook = True

        self.db.rep_factor = 1
        pool = round(self.db.active_masters / self.db.rep_factor)
        self.assertEqual(self.db.rep_pool_sz(), pool)

    def test_build_write_list_returns_all_mns_when_jump_idx_0(self):
        mns = PhoneBook.masternodes

        self.assertEqual(mns, self.db.build_wr_list(None, 0))

    def test_build_write_list_curr_node_0_jump_idx_1_returns_all(self):
        masternodes = list(range(100))
        delegates = list(range(10))
        big_vkbook = VKBook(masternodes, delegates, stamps=True, nonces=True, debug=True)

        self.db.vkbook = big_vkbook

        write_list = self.db.build_wr_list(0, 1)
        self.assertEqual(masternodes, write_list)

    def test_build_write_list_curr_node_20_jump_idx_1_returns_subset(self):
        masternodes = list(range(100))
        delegates = list(range(10))
        big_vkbook = VKBook(masternodes, delegates, stamps=True, nonces=True, debug=True)

        self.db.vkbook = big_vkbook

        write_list = self.db.build_wr_list(20, 1)
        self.assertEqual(masternodes[20:], write_list)

    def test_update_index(self):
        block = {
            'blockNum': 100,
            'blockHash': 'a',
            'data': 'woohoo'
        }

        owners = ['tejas', 'stu']

        self.db.update_index(block, owners)

        stored_index = self.db.get_index(100)

        self.assertEqual(stored_index['blockOwners'], owners)

    def test_update_index_no_owners(self):
        block = {
            'blockNum': 100,
            'blockHash': 'a',
            'data': 'woohoo'
        }

        with self.assertRaises(AssertionError):
            self.db.update_index(block, [])

    def test_update_index_no_block_hash(self):
        block = {
            'blockNum': 100,
            'data': 'woohoo'
        }

        with self.assertRaises(AssertionError):
            self.db.update_index(block, ['tejas', 'stu'])

    def test_update_index_fails_with_no_block_num(self):
        block = {
            'blockHash': 'a',
            'data': 'woohoo'
        }

        with self.assertRaises(AssertionError):
            self.db.update_index(block, ['tejas', 'stu'])

    def test_build_write_list_jump_idx_2_skips(self):
        masternodes = list(range(100))
        delegates = list(range(10))
        big_vkbook = VKBook(masternodes, delegates, stamps=True, nonces=True, debug=True)

        self.db.vkbook = big_vkbook

        write_list = self.db.build_wr_list(20, 2)
        self.assertEqual(masternodes[20::2], write_list)

    def test_evaluate_write_no_entry_returns_false(self):
        self.assertFalse(self.db.evaluate_wr())

    def test_evaluate_write_always_write_if_too_few_masters(self):
        self.db.active_masters = 1
        self.db.quorum_needed = 4

        block = {
            'blockNum': 103,
            'blockHash': 'a',
            'data': 'woohoo'
        }

        self.db.evaluate_wr(block)

        stored_block = self.db.get_block(103)
        block.pop('_id')

        self.assertEqual(stored_block, block)

        stored_index = self.db.get_index(103)

        owners = self.db.build_wr_list(self.db.mn_id, 0)

        self.assertEqual(stored_index['blockOwners'], owners)
        self.assertEqual(stored_index['blockHash'], block['blockHash'])
        self.assertEqual(stored_index['blockNum'], block['blockNum'])

    def test_eval_write_node_id_is_in_writers_returns_true(self):
        self.db.rep_factor = 1

        block = {
            'blockNum': 100,
            'blockHash': 'a',
            'data': 'woohoo'
        }

        res = self.db.evaluate_wr(block, node_id=0)

        self.assertTrue(res)

    def test_eval_write_node_id_is_not_in_writers_returns_false(self):
        self.db.rep_factor = 1

        block = {
            'blockNum': 100,
            'blockHash': 'a',
            'data': 'woohoo'
        }

        res = self.db.evaluate_wr(block, node_id=1)

        self.assertFalse(res)

    def test_eval_write_if_mn_is_writer_then_write_block(self):
        self.db.rep_factor = 1

        block = {
            'blockNum': 100,
            'blockHash': 'a',
            'data': 'woohoo'
        }

        self.db.evaluate_wr(block)

        stored_block = self.db.get_block(100)
        stored_index = self.db.get_index(100)

        block.pop('_id')

        self.assertEqual(block, stored_block)

        self.assertEqual(stored_index['blockHash'], block['blockHash'])
        self.assertEqual(stored_index['blockNum'], block['blockNum'])

    def test_get_full_block_by_number(self):

        block = {
            'blockNum': 100,
            'blockHash': 'a',
            'data': 'woohoo'
        }

        self.db.put(block, DistributedMasterStorage.BLOCK)

        block.pop('_id')

        stored_block = self.db.get_block(100)

        self.assertEqual(block, stored_block)

    def test_get_full_block_by_hash(self):
        block = {
            'blockNum': 101,
            'blockHash': 'abcdef',
            'data': 'woohoo'
        }

        self.db.blocks.collection.insert_one(block)

        block.pop('_id')

        stored_block = self.db.get_block('abcdef')

        self.assertEqual(block, stored_block)

    def test_get_full_block_returns_none_if_nothing_provided(self):
        self.assertIsNone(self.db.get_block())



from cilantro_ee.messages.transaction.data import TransactionDataBuilder
from cilantro_ee.messages.block_data.block_data import GENESIS_BLOCK_HASH
from cilantro_ee.core.containers.merkle_tree import MerkleTree
from cilantro_ee.messages.block_data.sub_block import SubBlock, SubBlockBuilder
from cilantro_ee.messages.consensus.merkle_signature import MerkleSignature

from unittest import TestCase
from cilantro_ee.constants.testnet import TESTNET_MASTERNODES
TEST_SK, TEST_VK = TESTNET_MASTERNODES[0]['sk'], TESTNET_MASTERNODES[0]['vk']


class TestSubBlock(TestCase):

    def test_create(self):
        txs = []
        for _ in range(8):
            txs.append(TransactionDataBuilder.create_random_tx())
        txs_bin = [tx.serialize() for tx in txs]

        tree = MerkleTree.from_raw_transactions(txs_bin)
        merkle_root = tree.root_as_hex
        input_hash = 'A'*64
        sb_idx = 0

        sk1, sk2 = 'AB' * 32, 'BC' * 32
        sigs = [MerkleSignature.create_from_payload(sk1, tree.root),
                MerkleSignature.create_from_payload(sk2, tree.root)]

        sb = SubBlock.create(merkle_root=merkle_root, signatures=sigs, merkle_leaves=tree.leaves_as_hex,
                             sub_block_idx=sb_idx, input_hash=input_hash, transactions=txs)

        self.assertEqual(sb.merkle_root, merkle_root)
        self.assertEqual(sb.signatures, sigs)
        self.assertEqual(sb.merkle_leaves, tree.leaves_as_hex)
        self.assertEqual(sb.input_hash, input_hash)
        self.assertEqual(sb.index, sb_idx)
        self.assertEqual(sb.transactions, txs)

    def test_create_no_transactions(self):
        txs = []
        for _ in range(8):
            txs.append(TransactionDataBuilder.create_random_tx())
        txs_bin = [tx.serialize() for tx in txs]

        tree = MerkleTree.from_raw_transactions(txs_bin)
        merkle_root = tree.root_as_hex
        input_hash = 'A'*64
        sb_idx = 0

        sk1, sk2 = 'AB' * 32, 'BC' * 32
        sigs = [MerkleSignature.create_from_payload(sk1, tree.root),
                MerkleSignature.create_from_payload(sk2, tree.root)]

        sb = SubBlock.create(merkle_root=merkle_root, signatures=sigs, merkle_leaves=tree.leaves_as_hex,
                             sub_block_idx=sb_idx, input_hash=input_hash, transactions=[])

        self.assertEqual(sb.merkle_root, merkle_root)
        self.assertEqual(sb.signatures, sigs)
        self.assertEqual(sb.merkle_leaves, tree.leaves_as_hex)
        self.assertEqual(sb.input_hash, input_hash)
        self.assertEqual(sb.index, sb_idx)
        self.assertEqual(sb.transactions, [])

    def test_serialize_deserialize(self):
        sb = SubBlockBuilder.create()
        clone = SubBlock.from_bytes(sb.serialize())
        self.assertEqual(sb, clone)

    def test_remove_tx_data(self):
        sb = SubBlockBuilder.create()

        self.assertTrue(len(sb.transactions) > 0)
        self.assertTrue(len(sb.merkle_leaves) > 0)

        sb.remove_tx_data()

        self.assertTrue(len(sb.transactions) == 0)
        self.assertTrue(len(sb.merkle_leaves) == 0)

    def test_remove_tx_data_from_clone(self):
        sb = SubBlockBuilder.create()
        clone = SubBlock.from_bytes(sb.serialize())

        self.assertTrue(len(clone.transactions) > 0)
        self.assertTrue(len(clone.merkle_leaves) > 0)

        clone.remove_tx_data()

        self.assertTrue(len(clone.transactions) == 0)
        self.assertTrue(len(clone.merkle_leaves) == 0)

    def test_sub_block_builder(self):
        # This shoudl not raise an error
        sb = SubBlockBuilder.create()




# constants for transaction batcher

# in secs
BATCHER_SLEEP_INTERVAL = 1
MAX_TXN_SUBMISSION_DELAY = 5

# TODO determine optimum size based on communication and storage efficiencies
MAX_TXNS_PER_SUB_BLOCK = 100




INTRODUCE_MOTION = 'introduce_motion'
VOTE_ON_MOTION = 'vote_on_motion'

NO_MOTION = 0
ADD_MASTER = 1
REMOVE_MASTER = 2
ADD_SEAT = 3
REMOVE_SEAT = 4

VOTING_PERIOD = datetime.DAYS * 1

S = Hash()


@construct
def seed(initial_masternodes, initial_open_seats):
    S['masternodes'] = initial_masternodes
    S['open_seats'] = initial_open_seats

    S['yays'] = 0
    S['nays'] = 0

    S['current_motion'] = NO_MOTION
    S['motion_opened'] = now


@export
def current_value():
    return {
        'masternodes': S['masternodes'],
        'open_seats': S['open_seats']
    }


@export
def vote(vk, obj):
    assert type(obj) == tuple, 'Pass a tuple!'

    arg = None
    try:
        action, position, arg = obj
    except ValueError:
        action, position = obj

    assert_vote_is_valid(vk, action, position, arg)

    if action == INTRODUCE_MOTION:
        introduce_motion(position, arg)

    else:
        assert S['current_motion'] != NO_MOTION, 'No motion proposed.'
        assert S['positions', vk] is None, 'VK already voted.'

        if position is True:
            S['yays'] += 1
            S['positions', vk] = position
        else:
            S['nays'] += 1
            S['positions', vk] = position

        if S['yays'] >= len(S['masternodes']) // 2 + 1:
            pass_current_motion()
            reset()

        elif S['nays'] >= len(S['masternodes']) // 2 + 1:
            reset()

        elif now - S['motion_opened'] >= VOTING_PERIOD:
            reset()


def assert_vote_is_valid(vk, action, position, arg=None):
    assert vk in S['masternodes'], 'Not a masternode.'

    assert action in [INTRODUCE_MOTION, VOTE_ON_MOTION], 'Invalid action.'

    if action == INTRODUCE_MOTION:
        assert S['current_motion'] == NO_MOTION, 'Already in motion.'
        assert 0 < position <= REMOVE_SEAT, 'Invalid motion.'
        if position == ADD_MASTER or position == REMOVE_MASTER:
            assert_vk_is_valid(arg)

    elif action == VOTE_ON_MOTION:
        assert type(position) == bool, 'Invalid position'


def assert_vk_is_valid(vk):
    assert vk is not None, 'No VK provided.'
    assert type(vk) == str, 'VK not a string.'
    assert len(vk) == 64, 'VK is not 64 characters.'
    int(vk, 16)


def introduce_motion(position, arg):
    if position == ADD_MASTER or position == REMOVE_SEAT:
        assert S['open_seats'] > 0, 'No open seats to add or remove.'

    if position == ADD_MASTER or position == REMOVE_MASTER:
        # If remove master, must be a master that already exists
        if position == REMOVE_MASTER:
            assert arg in S['masternodes'], 'Master does not exist.'

        S['master_in_question'] = arg

    S['current_motion'] = position
    S['motion_opened'] = now


def pass_current_motion():
    current_motion = S['current_motion']
    masters = S['masternodes']

    if current_motion == ADD_MASTER:
        masters.append(S['master_in_question'])
        S['open_seats'] -= 1

    elif current_motion == REMOVE_MASTER:
        masters.remove(S['master_in_question'])
        S['open_seats'] += 1

    elif current_motion == ADD_SEAT:
        S['open_seats'] += 1

    elif current_motion == REMOVE_SEAT:
        S['open_seats'] -= 1

    S['masternodes'] = masters


def reset():
    S['current_motion'] = NO_MOTION
    S['master_in_question'] = None
    S['yays'] = 0
    S['nays'] = 0
    S.clear('positions')



from cilantro_ee.utils.lprocess import LProcess
from cilantro_ee.nodes.base import NodeBase
from cilantro_ee.constants.system_config import NUM_SB_BUILDERS
from cilantro_ee.nodes.delegate.sub_block_builder_manager import SubBlockBuilderManager


class Delegate(NodeBase):

    # This call should not block!
    def start_node(self):
        self._start_bm()
        return NUM_SB_BUILDERS + 1

    def _start_bm(self):
        self.log.info("Starting sub-block builder manager process")
        self.bm = LProcess(target=SubBlockBuilderManager,
                           name='SubBlockBuilderManager',
                           kwargs={'signing_key': self.signing_key, 'ip': self.ip})
        self.bm.start()



from cilantro_ee.core.logger.base import get_logger
from multiprocessing import Process
import traceback, os, cProfile, pkg_resources, json
from vprof import runner


DELIM_LEN = 60
OUTER_DELIM = '!'
INNER_DELIM = '-'


class LProcess(Process):
    def run(self):
        log = get_logger(self.name)
        log.info("---> {} Starting --->".format(self.name))
        profiling = os.getenv('PROFILING', '')
        vprof_options = set('cmph').intersection(set(profiling))
        testname = os.getenv('TEST_NAME', 'sample_test')
        testid = os.getenv('TEST_ID', '')
        nodename = 'node'# os.getenv('HOST_NAME', 'node')
        profdir = os.path.join('profiles', testname, testid, nodename)
        profpath = os.path.join(profdir, self.name)
        try:
            # First create this directory for profiling
            if not os.path.exists(profdir) and profiling:
                os.makedirs(profdir)
            # Profiling with gprof to create Node graph profile
            if 'n' in profiling:
                pr = cProfile.Profile()
                pr.enable()
                super().run()
                pr.create_stats()
                pr.dump_stats('{}.stats'.format(profpath))
            # Profiling with vprof to create visualizations as specified in vprof_options
            elif len(vprof_options) > 0:
                vprof_options = ''.join(vprof_options)
                run_stats = runner.run_profilers((super().run, [], {}), vprof_options)
                with open('{}.json'.format(profpath), 'w+') as f:
                    run_stats['version'] = pkg_resources.get_distribution("vprof").version
                    f.write(json.dumps(run_stats))
            else:
                super().run()
        except Exception as e:
            err_msg = '\n' + OUTER_DELIM * DELIM_LEN
            err_msg += '\nException caught on ' + self.name + ':\n' + str(e)
            err_msg += '\n' + INNER_DELIM * DELIM_LEN
            err_msg += '\n' + traceback.format_exc()
            err_msg += '\n' + INNER_DELIM * DELIM_LEN
            err_msg += '\n' + OUTER_DELIM * DELIM_LEN
            log.error(err_msg)
        finally:
            log.info("<--- {} Terminating <---".format(self.name))
            # TODO -- signal to parent to call .join() on this process and clean up nicely



from seneca.libs.storage.datatypes import Hash

balances = Hash('balances', default_value=0)
custodials = Hash('custodials', default_value=0)


@seed
def seed():
    balances['324ee2e3544a8853a3c5a0ef0946b929aa488cbe7e7ee31a0fef9585ce398502'] = 1000000

@export
def transfer(to, amount):
    assert balances[rt['sender']] >= amount
    balances[to] += amount
    balances[rt['sender']] -= amount

@export
def add_to_custodial(to, amount):
    assert balances[rt['sender']] >= amount
    custodials[rt['sender']][to] += amount
    balances[rt['sender']] -= amount

@export
def remove_from_custodial(to, amount):
    assert custodials[rt['sender']][to] >= amount
    balances[rt['sender']] += amount
    custodials[rt['sender']][to] -= amount

@export
def spend_custodial(_from, amount, to):
    assert custodials[_from][rt['sender']] >= amount
    balances[to] += amount
    custodials[_from][rt['sender']] -= amount



from unittest import TestCase
from cilantro_ee.core.sockets import services
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.core.crypto import wallet
from cilantro_ee.services.block_server import BlockServer

from cilantro_ee.core.messages.message import Message
from cilantro_ee.core.messages.message_type import MessageType

from cilantro_ee.services.storage.master import CilantroStorageDriver
from cilantro_ee.core import canonical
from cilantro_ee.core.top import TopBlockManager
import time
import zmq.asyncio
import zmq
import asyncio
import hashlib
import secrets
from tests import random_txs

async def stop_server(s, timeout):
    await asyncio.sleep(timeout)
    s.stop()


class TestBlockServer(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.t = TopBlockManager()

    def tearDown(self):
        self.ctx.destroy()
        self.t.driver.flush()

    def test_get_latest_block_height(self):
        w = Wallet()
        m = BlockServer(services._socket('tcp://127.0.0.1:10000'), w, self.ctx, linger=500, poll_timeout=500)

        self.t.set_latest_block_number(555)

        async def get(msg):
            socket = self.ctx.socket(zmq.DEALER)
            socket.connect('tcp://127.0.0.1:10000')

            await socket.send(msg)

            res = await socket.recv()

            return res

        message = Message.get_signed_message_packed_2(wallet=w,
                                                      msg_type=MessageType.LATEST_BLOCK_HEIGHT_REQUEST,
                                                      timestamp=int(time.time()))

        tasks = asyncio.gather(
            m.serve(),
            get(message),
            stop_server(m, 0.2),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(res[1])

        self.assertEqual(msg.blockHeight, 555)

    def test_get_latest_block_hash(self):
        w = Wallet()
        m = BlockServer(services._socket('tcp://127.0.0.1:10000'), w, self.ctx, linger=500, poll_timeout=500)

        self.t.set_latest_block_hash(b'\xAA' * 32)

        async def get(msg):
            socket = self.ctx.socket(zmq.DEALER)
            socket.connect('tcp://127.0.0.1:10000')

            await socket.send(msg)

            res = await socket.recv()

            return res

        message = Message.get_signed_message_packed_2(wallet=w,
                                                      msg_type=MessageType.LATEST_BLOCK_HASH_REQUEST,
                                                      timestamp=int(time.time()))

        tasks = asyncio.gather(
            m.serve(),
            get(message),
            stop_server(m, 0.2),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(res[1])

        self.assertEqual(msg.blockHash, b'\xAA' * 32)

    def test_get_block_blob_by_block_data_request(self):
        block = random_txs.random_block()
        w = Wallet()
        c = CilantroStorageDriver(key=w.sk.encode().hex())
        c.drop_collections()

        d = canonical.block_from_subblocks([s for s in block.subBlocks], previous_hash=b'x/00' * 32, block_num=0)

        d['blockOwners'] = [secrets.token_bytes(32) for _ in range(12)]

        c.put(d)

        del d['_id']
        del d['blockOwners']

        m = BlockServer(services._socket('tcp://127.0.0.1:10000'), w, self.ctx, linger=2000, poll_timeout=500, driver=c)

        async def get(msg):
            socket = self.ctx.socket(zmq.DEALER)
            socket.connect('tcp://127.0.0.1:10000')

            await socket.send(msg)

            res = await socket.recv()

            return res

        message = Message.get_signed_message_packed_2(wallet=w,
                                                      msg_type=MessageType.BLOCK_DATA_REQUEST,
                                                      blockNum=0)

        tasks = asyncio.gather(
            m.serve(),
            get(message),
            stop_server(m, 0.2),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(res[1])

        dd = canonical.block_from_subblocks([s for s in msg.subBlocks], previous_hash=b'x/00' * 32, block_num=0)

        self.assertDictEqual(d, dd)

    def test_get_block_blob_by_block_but_failure_returns_bad_request(self):
        w = Wallet()
        c = CilantroStorageDriver(key=w.sk.encode().hex())
        c.drop_collections()
        m = BlockServer(services._socket('tcp://127.0.0.1:10000'), w, self.ctx, linger=500, poll_timeout=500, driver=c)

        async def get(msg):
            socket = self.ctx.socket(zmq.DEALER)
            socket.connect('tcp://127.0.0.1:10000')

            await socket.send(msg)

            res = await socket.recv()

            return res

        message = Message.get_signed_message_packed_2(wallet=w,
                                                      msg_type=MessageType.BLOCK_DATA_REQUEST,
                                                      blockNum=0)

        tasks = asyncio.gather(
            m.serve(),
            get(message),
            stop_server(m, 0.2),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(res[1])
        self.assertEqual(msg_type, MessageType.BAD_REQUEST)



# from cilantro_ee.core.logger.base import get_logger
# from cilantro_ee.core.utils.worker import Worker
#
# from cilantro_ee.utils.lprocess import LProcess
#
# from cilantro_ee.services.storage.vkbook import PhoneBook
#
# from cilantro_ee.nodes.catchup import CatchupManager
# from cilantro_ee.nodes.base import NodeBase
#
# from cilantro_ee.constants.ports import *
# from cilantro_ee.constants.zmq_filters import *
#
# from cilantro_ee.messages.envelope.envelope import Envelope
# from cilantro_ee.messages.block_data.notification import NewBlockNotification
# from cilantro_ee.messages.block_data.state_update import *
# from cilantro_ee.messages.signals.state_sync import UpdatedStateSignal
#
# import asyncio, zmq, math
# from collections import defaultdict
#
#
# IPC_ROUTER_IP = 'state-sync-router-ipc-sock'
# IPC_ROUTER_PORT = 6174
#
# IPC_PUB_IP = 'state-sync-pub-ipc-sock'
# IPC_PUB_PORT = 6175
#
#
# class StateSyncNode(NodeBase):
#     def start_node(self):
#         self.log.info("Starting StateSync processes")
#         self.sync = LProcess(target=StateSync, name='StateSync',
#                              kwargs={'signing_key': self.signing_key, 'ip': self.ip})
#         self.sync.start()
#
#
# class StateSync(Worker):
#     def __init__(self, ip, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.log = get_logger("StateSync")
#         self.ip = ip
#
#         # these guys get set in build_task_list
#         self.router, self.sub, self.pub, self.ipc_router, self.ipc_pub = None, None, None, None, None
#         self.cm = None
#
#         self.is_caught_up = False
#         self.new_blk_notifs = defaultdict(list)  # mapping of block hash to new_block_notif messages
#         self.mn_quorum = PhoneBook.masternode_quorum_min
#
#         self.run()
#
#     def run(self):
#         self.build_task_list()
#         self.log.info("StateSync starting event loop")
#         self.loop.run_until_complete(asyncio.gather(*self.tasks))
#
#     def build_task_list(self):
#         self.router = self.manager.create_socket(
#             socket_type=zmq.ROUTER,
#             name="StateSync-Router-{}".format(self.verifying_key[-4:]),
#             secure=True,
#         )
#         # self.router.setsockopt(zmq.ROUTER_MANDATORY, 1)  # FOR DEBUG ONLY
#         self.router.setsockopt(zmq.IDENTITY, self.verifying_key.encode())
#         self.router.bind(port=SS_ROUTER_PORT, protocol='tcp', ip=self.ip)
#         self.tasks.append(self.router.add_handler(self.handle_router_msg))
#
#         self.ipc_router = self.manager.create_socket(socket_type=zmq.ROUTER, name="BM-IPC-Router")
#         self.ipc_router.setsockopt(zmq.ROUTER_MANDATORY, 1)  # FOR DEBUG ONLY
#         self.ipc_router.bind(port=IPC_ROUTER_PORT, protocol='ipc', ip=IPC_ROUTER_IP)
#         self.tasks.append(self.ipc_router.add_handler(self.handle_ipc_msg))
#
#         self.ipc_pub = self.manager.create_socket(socket_type=zmq.PUB, name="StateSync-IPC-Pub", secure=False)
#         self.ipc_pub.bind(port=IPC_PUB_PORT, protocol="ipc", ip=IPC_PUB_IP)
#
#         self.pub = self.manager.create_socket(
#             socket_type=zmq.PUB,
#             name="StateSync-Pub-{}".format(self.verifying_key[-4:]),
#             secure=True,
#         )
#         self.pub.bind(port=SS_PUB_PORT, protocol='tcp', ip=self.ip)
#
#         self.sub = self.manager.create_socket(
#             socket_type=zmq.SUB,
#             name="StateSync-Sub-{}".format(self.verifying_key[-4:]),
#             secure=True,
#         )
#         self.sub.setsockopt(zmq.SUBSCRIBE, NEW_BLK_NOTIF_FILTER.encode())
#         self.tasks.append(self.sub.add_handler(self.handle_sub_msg))
#
#         self.tasks.append(self._connect_and_process())
#
#         self.cm = CatchupManager(verifying_key=self.verifying_key, signing_key=self.signing_key, pub_socket=self.pub, router_socket=self.router,
#                                  store_full_blocks=False)
#
#     async def catchup_db_state(self):
#         await asyncio.sleep(6)  # so pub/sub connections can complete
#         assert self.cm, "Expected catchup_mgr initialized at this point, learn to fking code pls"
#
#         self.log.info("Catching up...")
#         self.cm.run_catchup()
#
#     def check_catchup(self, caught_up: bool):
#         if not caught_up:
#             self.log.important("Setting caught up to false")
#             self.is_caught_up = caught_up
#
#         elif not self.is_caught_up and caught_up:
#             self.log.important("Setting caught up to true, and running caughtup behavior")
#             self.is_caught_up = True
#             self.new_blk_notifs.clear()
#             self.run_caughtup_behavior()
#
#     def run_caughtup_behavior(self):
#         self._send_msg_over_ipc_pub(UpdatedStateSignal.create())
#
#     async def _connect_and_process(self):
#         # first make sure, we have overlay server ready
#         await self._wait_until_ready()
#
#         # Listen to Masternodes over sub and connect router for cm communication
#         for vk in PhoneBook.masternodes:
#             self.sub.connect(vk=vk, port=MN_PUB_PORT)
#             self.router.connect(vk=vk, port=MN_ROUTER_PORT)
#
#         # now start the cm
#         await self.catchup_db_state()
#
#     def handle_sub_msg(self, frames):
#         # Unpack the frames
#         msg_filter, msg_type, msg_blob = frames
#
#         # TODO -- consensus on NewBlockNotifications before we feed it to the cm manager
#
#         if msg_type == MessageTypes.BLOCK_NOTIFICATION:
#             # Unpack the message
#             external_message = signal_capnp.ExternalMessage.from_bytes_packed(msg_blob)
#
#             # If the sender has signed the payload, continue
#             if not _verify(external_message.sender, external_message.data, external_message.signature):
#                 return
#
#             # Unpack the block
#             block = BlockNotification.unpack_block_notification(external_message.data)
#             if block.type.which() == "newBlock":
#                 self.log.info("Got NewBlockNotification from sender {} with hash {}".format(external_message.sender, block.blockHash.hex()))
#                 self._handle_new_blk_notif(block)
#         else:
#             self.log.warning("Got unexpected message type {}".format(type(msg)))
#
#     def _handle_new_blk_notif(self, nbc: notification_capnp.BlockNotification):
#         self.new_blk_notifs[nbc.block_hash].append(nbc)
#         if len(self.new_blk_notifs[nbc.block_hash]) >= self.mn_quorum:
#             del self.new_blk_notifs[nbc.block_hash]
#             self.log.important3("Sending NBC {} to catchup manager".format(nbc))  # TODO delete
#             self.check_catchup(self.cm.recv_new_blk_notif(nbc))
#
#     def handle_router_msg(self, frames):
#         envelope = Envelope.from_bytes(frames[-1])
#         sender = envelope.sender
#         assert sender.encode() == frames[0], "Sender vk {} does not match id frame {}".format(sender.encode(),
#                                                                                               frames[0])
#         msg = envelope.message
#
#         if isinstance(msg, BlockIndexReply):
#             self.log.debugv("Got BlockIndexReply {}".format(msg))
#             self.check_catchup(self.cm.recv_block_idx_reply(sender, msg))
#
#         elif isinstance(msg, BlockDataReply):
#             self.log.debugv("Got BlockDataReply {}".format(msg))
#             self.check_catchup(self.cm.recv_block_data_reply(msg))
#
#         else:
#             raise Exception("Got message type {} from ROUTER socket that it does not know how to handle"
#                             .format(type(msg)))
#
#     def handle_ipc_msg(self, frames):
#         self.log.spam("Got msg over ROUTER IPC from a SBB with frames: {}".format(frames))  # TODO delete this
#         assert len(frames) == 3, "Expected 3 frames: (id, msg_type, msg_blob). Got {} instead.".format(frames)
#
#         # TODO logic here
#
#     def _send_msg_over_ipc_pub(self, message: MessageBase):
#         self.log.debug("Publishing message {} over IPC".format(message))
#         message_type = MessageBase.registry[type(message)]
#         self.ipc_pub.send_multipart([STATESYNC_FILTER.encode(), message_type, message.serialize()])
#
#     def _send_msg_over_ipc_router(self, task_idx: int, message: MessageBase):
#         """ Convenience method to send a MessageBase instance over IPC router socket to a particular dealer.
#          Includes a frame to identify the type of message """
#         self.log.spam("Sending msg to task_idx {} with payload {}".format(task_idx, message))
#         assert isinstance(message, MessageBase), "Must pass in a MessageBase instance"
#         id_frame = str(task_idx).encode()
#         message_type = MessageBase.registry[type(message)]  # this is an int (enum) denoting the class of message
#         self.ipc_router.send_multipart([id_frame, message_type, message.serialize()])



from zmq.utils.z85 import encode
from nacl.signing import SigningKey, VerifyKey
from nacl.bindings import crypto_sign_ed25519_sk_to_curve25519
import hashlib


class Keys:
    is_setup = False

    @classmethod
    def setup(cls, sk_hex):
        if not cls.is_setup:
            nacl_sk = SigningKey(seed=bytes.fromhex(sk_hex))
            cls.sk = sk_hex
            cls.vk = nacl_sk.verify_key.encode().hex()

            # DEBUG -- TODO DELETE
            from cilantro_ee.core.crypto.wallet import get_vk
            wall_vk = get_vk(cls.sk)
            assert wall_vk == cls.vk, "wallet vk {} not match nacl vk {}".format(wall_vk, cls.vk)
            # END DEBUG

            cls.public_key = cls.vk2pk(cls.vk)
            cls.private_key = crypto_sign_ed25519_sk_to_curve25519(nacl_sk._signing_key)
            cls.is_setup = True

    #TODO replace with Wallet class that has the same functionality
    @staticmethod
    def vk2pk(vk):
        return encode(VerifyKey(bytes.fromhex(vk)).to_curve25519_public_key()._public_key)

    #TODO Deprecate
    @staticmethod
    def digest(s):
        if not isinstance(s, bytes):
            s = str(s).encode('utf8')
        return hashlib.sha1(s).digest()




from unittest import TestCase
# from cilantro_ee.messages.message import Serializer, signal_capnp, envelope_capnp
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.core.utils.pow import SHA3POWBytes

import capnp

from cilantro_ee.messages import capnp as schemas

transaction_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/transaction.capnp')


class TestSerializer(TestCase):
    def test_init(self):
        Serializer(capnp_type=signal_capnp.Signal)

    def test_unpack_returns_equal_capnp_struct(self):
        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        message = envelope_capnp.Message.new_message(payload=packed)
        packed_message = message.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal)

        unpacked_signal = serializer.unpack(packed_message)

        self.assertEqual(some_signal.messageType, unpacked_signal.messageType)

    def test_pack_returns_same_bytes_as_bytes_packed(self):
        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        message = envelope_capnp.Message.new_message(payload=packed)
        packed_message = message.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal)
        serialized = serializer.pack(packed)

        self.assertEqual(packed_message, serialized)

    def test_unpack_signed_improperly(self):
        w1 = Wallet()
        w2 = Wallet()

        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        message = envelope_capnp.Message.new_message(payload=packed)
        message.verifyingKey = w1.verifying_key()
        message.signature = w2.sign(message.payload)

        packed_message = message.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal, sign=True)

        unpacked = serializer.unpack(packed_message)

        self.assertIsNone(unpacked)

    def test_unpack_signed_properly(self):
        w1 = Wallet()

        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        message = envelope_capnp.Message.new_message(payload=packed)
        message.verifyingKey = w1.verifying_key()
        message.signature = w1.sign(message.payload)

        packed_message = message.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal, sign=True)

        unpacked = serializer.unpack(packed_message)

        self.assertEqual(unpacked.messageType, some_signal.messageType)

    def test_unpack_proof_incorrect(self):
        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        message = envelope_capnp.Message.new_message(payload=packed)
        message.proof = b'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'

        packed_message = message.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal, prove=True)

        unpacked = serializer.unpack(packed_message)

        self.assertIsNone(unpacked)

    def test_unpack_proof_correct(self):
        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        message = envelope_capnp.Message.new_message(payload=packed)
        message.proof = SHA3POWBytes.find(packed)

        packed_message = message.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal, prove=True)

        unpacked = serializer.unpack(packed_message)

        self.assertEqual(unpacked.messageType, some_signal.messageType)

    def test_pack_signs_properly(self):
        w1 = Wallet()

        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        message = envelope_capnp.Message.new_message(payload=packed)
        message.verifyingKey = w1.verifying_key()
        message.signature = w1.sign(message.payload)

        packed_message = message.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal, sign=True)

        packed_and_signed = serializer.pack(packed, w1)

        self.assertEqual(packed_message, packed_and_signed)

    def test_pack_signs_returns_none_if_no_wallet_provided(self):
        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal, sign=True)

        packed_and_signed = serializer.pack(packed)

        self.assertIsNone(packed_and_signed)

    def test_pack_proves_properly(self):
        some_signal = signal_capnp.Signal.new_message(messageType=123)
        packed = some_signal.to_bytes_packed()

        serializer = Serializer(capnp_type=signal_capnp.Signal, prove=True)

        packed_and_proven = serializer.pack(packed)

        message = envelope_capnp.Message.from_bytes_packed(packed_and_proven)

        self.assertTrue(SHA3POWBytes.check(message.payload, message.proof))
